import collections
import math
from functools import lru_cache
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterable,
    Iterator,
    Mapping,
    Sequence,
    TypeVar,
    Union,
)

from pip._vendor.resolvelib.providers import AbstractProvider

from .base import Candidate, Constraint, Requirement
from .candidates import REQUIRES_PYTHON_IDENTIFIER
from .factory import Factory

if TYPE_CHECKING:
    from pip._vendor.resolvelib.providers import Preference
    from pip._vendor.resolvelib.resolvers import RequirementInformation

    PreferenceInformation = RequirementInformation[Requirement, Candidate]

    _ProviderBase = AbstractProvider[Requirement, Candidate, str]
else:
    _ProviderBase = AbstractProvider

# Notes on the relationship between the provider, the factory, and the
# candidate and requirement classes.
#
# The provider is a direct implementation of the resolvelib class. Its role
# is to deliver the API that resolvelib expects.
#
# Rather than work with completely abstract "requirement" and "candidate"
# concepts as resolvelib does, pip has concrete classes implementing these two
# ideas. The API of Requirement and Candidate objects are defined in the base
# classes, but essentially map fairly directly to the equivalent provider
# methods. In particular, `find_matches` and `is_satisfied_by` are
# requirement methods, and `get_dependencies` is a candidate method.
#
# The factory is the interface to pip's internal mechanisms. It is stateless,
# and is created by the resolver and held as a property of the provider. It is
# responsible for creating Requirement and Candidate objects, and provides
# services to those objects (access to pip's finder and preparer).


D = TypeVar("D")
V = TypeVar("V")


def _get_with_identifier(
    mapping: Mapping[str, V],
    identifier: str,
    default: D,
) -> Union[D, V]:
    """Get item from a package name lookup mapping with a resolver identifier.

    This extra logic is needed when the target mapping is keyed by package
    name, which cannot be directly looked up with an identifier (which may
    contain requested extras). Additional logic is added to also look up a value
    by "cleaning up" the extras from the identifier.
    """
    if identifier in mapping:
        return mapping[identifier]
    # HACK: Theoretically we should check whether this identifier is a valid
    # "NAME[EXTRAS]" format, and parse out the name part with packaging or
    # some regular expression. But since pip's resolver only spits out three
    # kinds of identifiers: normalized PEP 503 names, normalized names plus
    # extras, and Requires-Python, we can cheat a bit here.
    name, open_bracket, _ = identifier.partition("[")
    if open_bracket and name in mapping:
        return mapping[name]
    return default


class PipProvider(_ProviderBase):
    """Pip's provider implementation for resolvelib.

    :params constraints: A mapping of constraints specified by the user. Keys
        are canonicalized project names.
    :params ignore_dependencies: Whether the user specified ``--no-deps``.
    :params upgrade_strategy: The user-specified upgrade strategy.
    :params user_requested: A set of canonicalized package names that the user
        supplied for pip to install/upgrade.
    """

    def __init__(
        self,
        factory: Factory,
        constraints: Dict[str, Constraint],
        ignore_dependencies: bool,
        upgrade_strategy: str,
        user_requested: Dict[str, int],
    ) -> None:
        self._factory = factory
        self._constraints = constraints
        self._ignore_dependencies = ignore_dependencies
        self._upgrade_strategy = upgrade_strategy
        self._user_requested = user_requested
        self._known_depths: Dict[str, float] = collections.defaultdict(lambda: math.inf)

    def identify(self, requirement_or_candidate: Union[Requirement, Candidate]) -> str:
        return requirement_or_candidate.name

    def get_preference(
        self,
        identifier: str,
        resolutions: Mapping[str, Candidate],
        candidates: Mapping[str, Iterator[Candidate]],
        information: Mapping[str, Iterable["PreferenceInformation"]],
        backtrack_causes: Sequence["PreferenceInformation"],
    ) -> "Preference":
        """Produce a sort key for given requirement based on preference.

        The lower the return value is, the more preferred this group of
        arguments is.

        Currently pip considers the following in order:

        * Prefer if any of the known requirements is "direct", e.g. points to an
          explicit URL.
        * If equal, prefer if any requirement is "pinned", i.e. contains
          operator ``===`` or ``==``.
        * If equal, calculate an approximate "depth" and resolve requirements
          closer to the user-specified requirements first. If the depth cannot
          by determined (eg: due to no matching parents), it is considered
          infinite.
        * Order user-specified requirements by the order they are specified.
        * If equal, prefers "non-free" requirements, i.e. contains at least one
          operator, such as ``>=`` or ``<``.
        * If equal, order alphabetically for consistency (helps debuggability).
        """
        try:
            next(iter(information[identifier]))
        except StopIteration:
            # There is no information for this identifier, so there's no known
            # candidates.
            has_information = False
        else:
            has_information = True

        if has_information:
            lookups = (r.get_candidate_lookup() for r, _ in information[identifier])
            candidate, ireqs = zip(*lookups)
        else:
            candidate, ireqs = None, ()

        operators = [
            specifier.operator
            for specifier_set in (ireq.specifier for ireq in ireqs if ireq)
            for specifier in specifier_set
        ]

        direct = candidate is not None
        pinned = any(op[:2] == "==" for op in operators)
        unfree = bool(operators)

        try:
            requested_order: Union[int, float] = self._user_requested[identifier]
        except KeyError:
            requested_order = math.inf
            if has_information:
                parent_depths = (
                    self._known_depths[parent.name] if parent is not None else 0.0
                    for _, parent in information[identifier]
                )
                inferred_depth = min(d for d in parent_depths) + 1.0
            else:
                inferred_depth = math.inf
        else:
            inferred_depth = 1.0
        self._known_depths[identifier] = inferred_depth

        requested_order = self._user_requested.get(identifier, math.inf)

        # Requires-Python has only one candidate and the check is basically
        # free, so we always do it first to avoid needless work if it fails.
        requires_python = identifier == REQUIRES_PYTHON_IDENTIFIER

        # Prefer the causes of backtracking on the assumption that the problem
        # resolving the dependency tree is related to the failures that caused
        # the backtracking
        backtrack_cause = self.is_backtrack_cause(identifier, backtrack_causes)

        return (
            not requires_python,
            not direct,
            not pinned,
            not backtrack_cause,
            inferred_depth,
            requested_order,
            not unfree,
            identifier,
        )

    def find_matches(
        self,
        identifier: str,
        requirements: Mapping[str, Iterator[Requirement]],
        incompatibilities: Mapping[str, Iterator[Candidate]],
    ) -> Iterable[Candidate]:
        def _eligible_for_upgrade(identifier: str) -> bool:
            """Are upgrades allowed for this project?

            This checks the upgrade strategy, and whether the project was one
            that the user specified in the command line, in order to decide
            whether we should upgrade if there's a newer version available.

            (Note that we don't need access to the `--upgrade` flag, because
            an upgrade strategy of "to-satisfy-only" means that `--upgrade`
            was not specified).
            """
            if self._upgrade_strategy == "eager":
                return True
            elif self._upgrade_strategy == "only-if-needed":
                user_order = _get_with_identifier(
                    self._user_requested,
                    identifier,
                    default=None,
                )
                return user_order is not None
            return False

        constraint = _get_with_identifier(
            self._constraints,
            identifier,
            default=Constraint.empty(),
        )
        return self._factory.find_candidates(
            identifier=identifier,
            requirements=requirements,
            constraint=constraint,
            prefers_installed=(not _eligible_for_upgrade(identifier)),
            incompatibilities=incompatibilities,
            is_satisfied_by=self.is_satisfied_by,
        )

    @lru_cache(maxsize=None)
    def is_satisfied_by(self, requirement: Requirement, candidate: Candidate) -> bool:
        return requirement.is_satisfied_by(candidate)

    def get_dependencies(self, candidate: Candidate) -> Sequence[Requirement]:
        with_requires = not self._ignore_dependencies
        return [r for r in candidate.iter_dependencies(with_requires) if r is not None]

    @staticmethod
    def is_backtrack_cause(
        identifier: str, backtrack_causes: Sequence["PreferenceInformation"]
    ) -> bool:
        for backtrack_cause in backtrack_causes:
            if identifier == backtrack_cause.requirement.name:
                return True
            if backtrack_cause.parent and identifier == backtrack_cause.parent.name:
                return True
        return False



"""ModbusMessage layer.

is extending ModbusProtocol to handle receiving and sending of messsagees.

ModbusMessage provides a unified interface to send/receive Modbus requests/responses.
"""
from __future__ import annotations

from binascii import a2b_hex, b2a_hex

from pymodbus.framer.base import FramerBase
from pymodbus.logging import Log


class FramerAscii(FramerBase):
    r"""Modbus ASCII Frame Controller.

    Layout::
        [ Start ][ Dev id ][ Function ][ Data ][ LRC ][ End ]
          1c       2c        2c          N*2c    1c     2c

        * data can be 1 - 2x252 chars
        * end is "\\r\\n" (Carriage return line feed), however the line feed
          character can be changed via a special command
        * start is ":"

    This framer is used for serial transmission.  Unlike the RTU protocol,
    the data in this framer is transferred in plain text ascii.
    """

    START = b':'
    END = b'\r\n'
    MIN_SIZE = 10


    def decode(self, data: bytes) -> tuple[int, int, int, bytes]:
        """Decode ADU."""
        used_len = 0
        data_len = len(data)
        while True:
            if data_len - used_len < self.MIN_SIZE:
                Log.debug("Short frame: {} wait for more data", data, ":hex")
                return used_len, 0, 0, self.EMPTY
            buffer = data[used_len:]
            if buffer[0:1] != self.START:
                if (i := buffer.find(self.START)) == -1:
                    Log.debug("No frame start in data: {}, wait for data", data, ":hex")
                    return data_len, 0, 0, self.EMPTY
                used_len += i
                continue
            if (end := buffer.find(self.END)) == -1:
                Log.debug("Incomplete frame: {} wait for more data", data, ":hex")
                return used_len, 0, 0, self.EMPTY
            dev_id = int(buffer[1:3], 16)
            lrc = int(buffer[end - 2: end], 16)
            msg = a2b_hex(buffer[1 : end - 2])
            used_len += end + 2
            if not self.check_LRC(msg, lrc):
                Log.debug("LRC wrong in frame: {} skipping", data, ":hex")
                continue
            return used_len, dev_id, 0, msg[1:]

    def encode(self, data: bytes, device_id: int, _tid: int) -> bytes:
        """Encode ADU."""
        dev_id = device_id.to_bytes(1,'big')
        checksum = self.compute_LRC(dev_id + data)
        frame = (
            self.START +
            f"{device_id:02x}".encode() +
            b2a_hex(data) +
            f"{checksum:02x}".encode() +
            self.END
        ).upper()
        return frame

    @classmethod
    def compute_LRC(cls, data: bytes) -> int:
        """Use to compute the longitudinal redundancy check against a string."""
        lrc = sum(int(a) for a in data) & 0xFF
        lrc = (lrc ^ 0xFF) + 1
        return lrc & 0xFF

    @classmethod
    def check_LRC(cls, data: bytes, check: int) -> bool:
        """Check if the passed in data matches the LRC."""
        return cls.compute_LRC(data) == check



from dataclasses import dataclass
from typing import FrozenSet, Iterable, Optional, Tuple

from pip._vendor.packaging.specifiers import SpecifierSet
from pip._vendor.packaging.utils import NormalizedName
from pip._vendor.packaging.version import Version

from pip._internal.models.link import Link, links_equivalent
from pip._internal.req.req_install import InstallRequirement
from pip._internal.utils.hashes import Hashes

CandidateLookup = Tuple[Optional["Candidate"], Optional[InstallRequirement]]


def format_name(project: NormalizedName, extras: FrozenSet[NormalizedName]) -> str:
    if not extras:
        return project
    extras_expr = ",".join(sorted(extras))
    return f"{project}[{extras_expr}]"


@dataclass(frozen=True)
class Constraint:
    specifier: SpecifierSet
    hashes: Hashes
    links: FrozenSet[Link]

    @classmethod
    def empty(cls) -> "Constraint":
        return Constraint(SpecifierSet(), Hashes(), frozenset())

    @classmethod
    def from_ireq(cls, ireq: InstallRequirement) -> "Constraint":
        links = frozenset([ireq.link]) if ireq.link else frozenset()
        return Constraint(ireq.specifier, ireq.hashes(trust_internet=False), links)

    def __bool__(self) -> bool:
        return bool(self.specifier) or bool(self.hashes) or bool(self.links)

    def __and__(self, other: InstallRequirement) -> "Constraint":
        if not isinstance(other, InstallRequirement):
            return NotImplemented
        specifier = self.specifier & other.specifier
        hashes = self.hashes & other.hashes(trust_internet=False)
        links = self.links
        if other.link:
            links = links.union([other.link])
        return Constraint(specifier, hashes, links)

    def is_satisfied_by(self, candidate: "Candidate") -> bool:
        # Reject if there are any mismatched URL constraints on this package.
        if self.links and not all(_match_link(link, candidate) for link in self.links):
            return False
        # We can safely always allow prereleases here since PackageFinder
        # already implements the prerelease logic, and would have filtered out
        # prerelease candidates if the user does not expect them.
        return self.specifier.contains(candidate.version, prereleases=True)


class Requirement:
    @property
    def project_name(self) -> NormalizedName:
        """The "project name" of a requirement.

        This is different from ``name`` if this requirement contains extras,
        in which case ``name`` would contain the ``[...]`` part, while this
        refers to the name of the project.
        """
        raise NotImplementedError("Subclass should override")

    @property
    def name(self) -> str:
        """The name identifying this requirement in the resolver.

        This is different from ``project_name`` if this requirement contains
        extras, where ``project_name`` would not contain the ``[...]`` part.
        """
        raise NotImplementedError("Subclass should override")

    def is_satisfied_by(self, candidate: "Candidate") -> bool:
        return False

    def get_candidate_lookup(self) -> CandidateLookup:
        raise NotImplementedError("Subclass should override")

    def format_for_error(self) -> str:
        raise NotImplementedError("Subclass should override")


def _match_link(link: Link, candidate: "Candidate") -> bool:
    if candidate.source_link:
        return links_equivalent(link, candidate.source_link)
    return False


class Candidate:
    @property
    def project_name(self) -> NormalizedName:
        """The "project name" of the candidate.

        This is different from ``name`` if this candidate contains extras,
        in which case ``name`` would contain the ``[...]`` part, while this
        refers to the name of the project.
        """
        raise NotImplementedError("Override in subclass")

    @property
    def name(self) -> str:
        """The name identifying this candidate in the resolver.

        This is different from ``project_name`` if this candidate contains
        extras, where ``project_name`` would not contain the ``[...]`` part.
        """
        raise NotImplementedError("Override in subclass")

    @property
    def version(self) -> Version:
        raise NotImplementedError("Override in subclass")

    @property
    def is_installed(self) -> bool:
        raise NotImplementedError("Override in subclass")

    @property
    def is_editable(self) -> bool:
        raise NotImplementedError("Override in subclass")

    @property
    def source_link(self) -> Optional[Link]:
        raise NotImplementedError("Override in subclass")

    def iter_dependencies(self, with_requires: bool) -> Iterable[Optional[Requirement]]:
        raise NotImplementedError("Override in subclass")

    def get_install_requirement(self) -> Optional[InstallRequirement]:
        raise NotImplementedError("Override in subclass")

    def format_for_error(self) -> str:
        raise NotImplementedError("Subclass should override")



"""Implementation of a Threaded Modbus Server."""
from __future__ import annotations

import asyncio
import os

from pymodbus.datastore import ModbusServerContext

from .base import ModbusBaseServer
from .server import (
    ModbusSerialServer,
    ModbusTcpServer,
    ModbusTlsServer,
    ModbusUdpServer,
)


async def StartAsyncTcpServer(
    context: ModbusServerContext,
    **kwargs,
) -> None:
    """Start and run a tcp modbus server.

    :parameter context: Datastore object
    :parameter kwargs: for parameter explanation see ModbusTcpServer

    .. tip::
        Only handles a single server !

        Use ModbusTcpServer to allow multiple servers in one app.
    """
    await ModbusTcpServer(context, **kwargs).serve_forever()


def StartTcpServer(
    context: ModbusServerContext,
    **kwargs
) -> None:
    """Start and run a modbus TCP server.

    :parameter context: Datastore object
    :parameter kwargs: for parameter explanation see ModbusTcpServer

    .. tip::
        Only handles a single server !

        Use ModbusTcpServer to allow multiple servers in one app.
    """
    asyncio.run(StartAsyncTcpServer(context, **kwargs))


async def StartAsyncTlsServer(
    context: ModbusServerContext,
    **kwargs,
) -> None:
    """Start and run a tls modbus server.

    :parameter context: Datastore object
    :parameter kwargs: for parameter explanation see ModbusTlsServer

    .. tip::
        Only handles a single server !

        Use ModbusTlsServer to allow multiple servers in one app.
    """
    await ModbusTlsServer(context, **kwargs).serve_forever()


def StartTlsServer(
    context: ModbusServerContext,
    **kwargs
) -> None:
    """Start and run a modbus TLS server.

    :parameter context: Datastore object
    :parameter kwargs: for parameter explanation see ModbusTlsServer

    .. tip::
        Only handles a single server !

        Use ModbusTlsServer to allow multiple servers in one app.
    """
    asyncio.run(StartAsyncTlsServer(context, **kwargs))


async def StartAsyncUdpServer(
    context: ModbusServerContext,
    **kwargs,
) -> None:
    """Start and run a udp modbus server.

    :parameter context: Datastore object
    :parameter kwargs: for parameter explanation see ModbusUdpServer

    .. tip::
        Only handles a single server !

        Use ModbusUdpServer to allow multiple servers in one app.
    """
    await ModbusUdpServer(context, **kwargs).serve_forever()


def StartUdpServer(
    context: ModbusServerContext,
    **kwargs
) -> None:
    """Start and run a modbus UDP server.

    :parameter context: Datastore object
    :parameter kwargs: for parameter explanation see ModbusUdpServer

    .. tip::
        Only handles a single server !

        Use ModbusUdpServer to allow multiple servers in one app.
    """
    asyncio.run(StartAsyncUdpServer(context, **kwargs))


async def StartAsyncSerialServer(
    context: ModbusServerContext,
    **kwargs,
) -> None:
    """Start and run a serial modbus server.

    :parameter context: Datastore object
    :parameter kwargs: for parameter explanation see ModbusSerialServer

    .. tip::
        Only handles a single server !

        Use ModbusSerialServer to allow multiple servers in one app.
    """
    await ModbusSerialServer(context, **kwargs).serve_forever()


def StartSerialServer(
    context: ModbusServerContext,
    **kwargs
) -> None:
    """Start and run a modbus serial server.

    :parameter context: Datastore object
    :parameter kwargs: for parameter explanation see ModbusSerialServer

    .. tip::
        Only handles a single server !

        Use ModbusSerialServer to allow multiple servers in one app.
    """
    asyncio.run(StartAsyncSerialServer(context, **kwargs))


async def ServerAsyncStop() -> None:
    """Terminate server."""
    if not ModbusBaseServer.active_server:
        raise RuntimeError("Modbus server not running.")
    await ModbusBaseServer.active_server.shutdown()
    ModbusBaseServer.active_server = None


def ServerStop() -> None:
    """Terminate server."""
    if not ModbusBaseServer.active_server:
        raise RuntimeError("Modbus server not running.")
    future = asyncio.run_coroutine_threadsafe(ServerAsyncStop(), ModbusBaseServer.active_server.loop)
    future.result(timeout=10 if os.name == 'nt' else 0.1)



"""HTTP server for modbus simulator."""
from __future__ import annotations

import asyncio
import contextlib
import dataclasses
import importlib
import json
import os
from typing import TYPE_CHECKING


try:
    from aiohttp import web

    AIOHTTP_MISSING = False
except ImportError:
    AIOHTTP_MISSING = True
    if TYPE_CHECKING:  # always False at runtime
        # type checkers do not understand the Raise RuntimeError in __init__()
        from aiohttp import web

from pymodbus.datastore import ModbusServerContext, ModbusSimulatorContext
from pymodbus.datastore.simulator import Label
from pymodbus.device import ModbusDeviceIdentification
from pymodbus.logging import Log
from pymodbus.pdu import DecodePDU
from pymodbus.server.server import (
    ModbusSerialServer,
    ModbusTcpServer,
    ModbusTlsServer,
    ModbusUdpServer,
)


MAX_FILTER = 1000

RESPONSE_INACTIVE = -1
RESPONSE_NORMAL = 0
RESPONSE_ERROR = 1
RESPONSE_EMPTY = 2
RESPONSE_JUNK = 3


@dataclasses.dataclass()
class CallTracer:
    """Define call/response traces."""

    call: bool = False
    fc: int = -1
    address: int = -1
    count: int = -1
    data: bytes = b""


@dataclasses.dataclass()
class CallTypeMonitor:
    """Define Request/Response monitor."""

    active: bool = False
    trace_response: bool = False
    range_start: int = -1
    range_stop: int = -1
    function: int = -1
    hex: bool = False
    decode: bool = False


@dataclasses.dataclass()
class CallTypeResponse:
    """Define Response manipulation."""

    active: int = RESPONSE_INACTIVE
    split: int = 0
    delay: int = 0
    junk_len: int = 10
    error_response: int = 0
    change_rate: int = 0
    clear_after: int = 1


class ModbusSimulatorServer:
    """**ModbusSimulatorServer**.

    :param modbus_server: Server name in json file (default: "server")
    :param modbus_device: Device name in json file (default: "client")
    :param http_host: TCP host for HTTP (default: "localhost")
    :param http_port: TCP port for HTTP (default: 8080)
    :param json_file: setup file (default: "setup.json")
    :param custom_actions_module: python module with custom actions (default: none)

    if either http_port or http_host is none, HTTP will not be started.
    This class starts a http server, that serves a couple of endpoints:

    - **"<addr>/"** static files
    - **"<addr>/api/log"** log handling, HTML with GET, REST-API with post
    - **"<addr>/api/registers"** register handling, HTML with GET, REST-API with post
    - **"<addr>/api/calls"** call (function code / message) handling, HTML with GET, REST-API with post
    - **"<addr>/api/server"** server handling, HTML with GET, REST-API with post

    Example::

        from pymodbus.server import ModbusSimulatorServer

        async def run():
            simulator = ModbusSimulatorServer(
                modbus_server="my server",
                modbus_device="my device",
                http_host="localhost",
                http_port=8080)
            await simulator.run_forever(only_start=True)
            ...
            await simulator.stop()
    """

    def __init__(
        self,
        modbus_server: str = "server",
        modbus_device: str = "device",
        http_host: str = "0.0.0.0",
        http_port: int = 8080,
        log_file: str = "server.log",
        json_file: str = "setup.json",
        custom_actions_module: str | None = None,
    ):
        """Initialize http interface."""
        if AIOHTTP_MISSING:
            raise RuntimeError(
                "Simulator server requires aiohttp. "
                'Please install with "pip install aiohttp" and try again.'
            )
        with open(json_file, encoding="utf-8") as file:
            setup = json.load(file)

        comm_class = {
            "serial": ModbusSerialServer,
            "tcp": ModbusTcpServer,
            "tls": ModbusTlsServer,
            "udp": ModbusUdpServer,
        }
        if custom_actions_module:
            actions_module = importlib.import_module(custom_actions_module)
            custom_actions_dict = actions_module.custom_actions_dict
        else:
            custom_actions_dict = {}
        server = setup["server_list"][modbus_server]
        if server["comm"] != "serial":
            server["address"] = (server["host"], server["port"])
            del server["host"]
            del server["port"]
        device = setup["device_list"][modbus_device]
        self.datastore_context = ModbusSimulatorContext(
            device, custom_actions_dict or {}
        )
        datastore = None
        if "device_id" in server:
            # Designated ModBus unit address. Will only serve data if the address matches
            datastore = ModbusServerContext(slaves={int(server["device_id"]): self.datastore_context}, single=False)
            del server["device_id"]
        else:
            # Will server any request regardless of addressing
            datastore = ModbusServerContext(slaves=self.datastore_context, single=True)

        comm = comm_class[server.pop("comm")]
        framer = server.pop("framer")
        if "identity" in server:
            server["identity"] = ModbusDeviceIdentification(
                info_name=server["identity"]
            )
        self.modbus_server = comm(framer=framer, context=datastore, **server)
        self.serving: asyncio.Future = asyncio.Future()
        self.log_file = log_file
        self.site: web.TCPSite | None = None
        self.runner: web.AppRunner
        self.http_host = http_host
        self.http_port = http_port
        self.web_path = os.path.join(os.path.dirname(__file__), "web")
        self.web_app = web.Application()
        self.web_app.add_routes(
            [
                web.get("/api/{tail:[a-z]*}", self.handle_html),
                web.post("/restapi/{tail:[a-z]*}", self.handle_json),
                web.get("/{tail:[a-z0-9.]*}", self.handle_html_static),
                web.get("/", self.handle_html_static),
            ]
        )
        self.web_app.on_startup.append(self.start_modbus_server)
        self.web_app.on_shutdown.append(self.stop_modbus_server)
        self.generator_html: dict[str, list] = {
            "log": ["", self.build_html_log],
            "registers": ["", self.build_html_registers],
            "calls": ["", self.build_html_calls],
            "server": ["", self.build_html_server],
        }
        self.generator_json = {
            "log": self.build_json_log,
            "registers": self.build_json_registers,
            "calls": self.build_json_calls,
            "server": self.build_json_server,
        }
        self.submit_html = {
            "Clear": self.action_clear,
            "Stop": self.action_stop,
            "Reset": self.action_reset,
            "Add": self.action_add,
            "Monitor": self.action_monitor,
            "Set": self.action_set,
            "Simulate": self.action_simulate,
        }
        for entry in self.generator_html:  # pylint: disable=consider-using-dict-items
            html_file = os.path.join(self.web_path, "generator", entry)
            with open(html_file, encoding="utf-8") as handle:
                self.generator_html[entry][0] = handle.read()
        self.refresh_rate = 0
        self.register_filter: list[int] = []
        self.call_list: list[CallTracer] = []
        self.request_lookup = DecodePDU(True).lookup
        self.call_monitor = CallTypeMonitor()
        self.call_response = CallTypeResponse()
        app_key = getattr(web, 'AppKey', str)  # fall back to str for aiohttp < 3.9.0
        self.api_key = app_key("modbus_server")

    async def start_modbus_server(self, app):
        """Start Modbus server as asyncio task."""
        try:
            if getattr(self.modbus_server, "start", None):
                await self.modbus_server.start()
            app[self.api_key] = asyncio.create_task(
                self.modbus_server.serve_forever()
            )
            app[self.api_key].set_name("simulator modbus server")
        except Exception as exc:
            Log.error("Error starting modbus server, reason: {}", exc)
            raise exc
        Log.info(
            "Modbus server started on {}", self.modbus_server.comm_params.source_address
        )

    async def stop_modbus_server(self, app):
        """Stop modbus server."""
        Log.info("Stopping modbus server")
        await self.modbus_server.shutdown()
        app[self.api_key].cancel()
        with contextlib.suppress(asyncio.exceptions.CancelledError):
            await app[self.api_key]

        Log.info("Modbus server Stopped")

    async def run_forever(self, only_start=False):
        """Start modbus and http servers."""
        try:
            self.runner = web.AppRunner(self.web_app)
            await self.runner.setup()
            self.site = web.TCPSite(self.runner, self.http_host, self.http_port)
            await self.site.start()
        except Exception as exc:
            Log.error("Error starting http server, reason: {}", exc)
            raise exc
        Log.info("HTTP server started on ({}:{})", self.http_host, self.http_port)
        if only_start:
            return
        await self.serving

    async def stop(self):
        """Stop modbus and http servers."""
        await self.runner.cleanup()
        self.site = None
        if not self.serving.done():
            self.serving.set_result(True)
        await asyncio.sleep(0)

    async def handle_html_static(self, request):
        """Handle static html."""
        if not (page := request.path[1:]):
            page = "index.html"
        file = os.path.normpath(os.path.join(self.web_path, page))
        if not file.startswith(self.web_path):
            raise ValueError(f"File access outside {self.web_path} not permitted.")
        try:
            with open(file, encoding="utf-8"):
                return web.FileResponse(file)
        except (FileNotFoundError, IsADirectoryError) as exc:
            raise web.HTTPNotFound(reason="File not found") from exc

    async def handle_html(self, request):
        """Handle html."""
        page_type = request.path.split("/")[-1]
        params = dict(request.query)
        if refresh := params.pop("refresh", None):
            self.refresh_rate = int(refresh)
        if self.refresh_rate > 0:
            html = self.generator_html[page_type][0].replace(
                "<!--REFRESH-->",
                f'<meta http-equiv="refresh" content="{self.refresh_rate}">',
            )
        else:
            html = self.generator_html[page_type][0].replace("<!--REFRESH-->", "")
        new_page = self.generator_html[page_type][1](params, html)
        return web.Response(text=new_page, content_type="text/html")

    async def handle_json(self, request):
        """Handle api registers."""
        command = request.path.split("/")[-1]
        params = await request.json()
        try:
            result = self.generator_json[command](params)
        except (KeyError, ValueError, TypeError, IndexError) as exc:
            Log.error("Unhandled error during json request: {}", exc)
            return web.json_response({"result": "error", "error": f"Unhandled error Error: {exc}"})
        return web.json_response(result)

    def build_html_registers(self, params, html):
        """Build html registers page."""
        result_txt, foot = self.helper_handle_submit(params, self.submit_html)
        if not result_txt:
            result_txt = "ok"
        if not foot:
            if self.register_filter:
                foot = f"{len(self.register_filter)} register(s) monitored"
            else:
                foot = "Nothing selected"
        register_types = "".join(
            f"<option value={reg_id}>{name}</option>"
            for name, reg_id in self.datastore_context.registerType_name_to_id.items()
        )
        register_actions = "".join(
            f"<option value={action_id}>{name}</option>"
            for name, action_id in self.datastore_context.action_name_to_id.items()
        )
        rows = ""
        for i in self.register_filter:
            inx, reg = self.datastore_context.get_text_register(i)
            if reg.type == Label.next:
                continue
            row = "".join(
                f"<td>{entry}</td>"
                for entry in (
                    inx,
                    reg.type,
                    reg.access,
                    reg.action,
                    reg.value,
                    reg.count_read,
                    reg.count_write,
                )
            )
            rows += f"<tr>{row}</tr>"
        new_html = (
            html.replace("<!--REGISTER_ACTIONS-->", register_actions)
            .replace("<!--REGISTER_TYPES-->", register_types)
            .replace("<!--REGISTER_FOOT-->", foot)
            .replace("<!--REGISTER_ROWS-->", rows)
            .replace("<!--RESULT-->", result_txt)
        )
        return new_html

    def build_html_calls(self, params: dict, html: str) -> str:
        """Build html calls page."""
        result_txt, foot = self.helper_handle_submit(params, self.submit_html)
        if not foot:
            foot = "Montitoring active" if self.call_monitor.active else "not active"
        if not result_txt:
            result_txt = "ok"

        function_error = ""
        for i, txt in (
            (1, "ILLEGAL_FUNCTION"),
            (2, "ILLEGAL_ADDRESS"),
            (3, "ILLEGAL_VALUE"),
            (4, "SLAVE_FAILURE"),
            (5, "ACKNOWLEDGE"),
            (6, "SLAVE_BUSY"),
            (7, "MEMORY_PARITY_ERROR"),
            (10, "GATEWAY_PATH_UNAVIABLE"),
            (11, "GATEWAY_NO_RESPONSE"),
        ):
            selected = "selected" if i == self.call_response.error_response else ""
            function_error += f"<option value={i} {selected}>{txt}</option>"
        range_start_html = (
            str(self.call_monitor.range_start)
            if self.call_monitor.range_start != -1
            else ""
        )
        range_stop_html = (
            str(self.call_monitor.range_stop)
            if self.call_monitor.range_stop != -1
            else ""
        )
        function_codes = ""
        for function in self.request_lookup.values():
            selected = (
                "selected"
                if function.function_code == self.call_monitor.function
                else ""
            )
            function_codes += f"<option value={function.function_code} {selected}>function code name</option>"
        simulation_action = (
            "ACTIVE" if self.call_response.active != RESPONSE_INACTIVE else ""
        )

        max_len = MAX_FILTER if self.call_monitor.active else 0
        while len(self.call_list) > max_len:
            del self.call_list[0]
        call_rows = ""
        for entry in reversed(self.call_list):
            # req_obj = self.request_lookup[entry[1]]
            call_rows += f"<tr><td>{entry.call} - {entry.fc}</td><td>{entry.address}</td><td>{entry.count}</td><td>{entry.data.decode()}</td></tr>"
            # line += req_obj.funcion_code_name
        new_html = (
            html.replace("<!--SIMULATION_ACTIVE-->", simulation_action)
            .replace("FUNCTION_RANGE_START", range_start_html)
            .replace("FUNCTION_RANGE_STOP", range_stop_html)
            .replace("<!--FUNCTION_CODES-->", function_codes)
            .replace(
                "FUNCTION_SHOW_HEX_CHECKED", "checked" if self.call_monitor.hex else ""
            )
            .replace(
                "FUNCTION_SHOW_DECODED_CHECKED",
                "checked" if self.call_monitor.decode else "",
            )
            .replace(
                "FUNCTION_RESPONSE_NORMAL_CHECKED",
                "checked" if self.call_response.active == RESPONSE_NORMAL else "",
            )
            .replace(
                "FUNCTION_RESPONSE_ERROR_CHECKED",
                "checked" if self.call_response.active == RESPONSE_ERROR else "",
            )
            .replace(
                "FUNCTION_RESPONSE_EMPTY_CHECKED",
                "checked" if self.call_response.active == RESPONSE_EMPTY else "",
            )
            .replace(
                "FUNCTION_RESPONSE_JUNK_CHECKED",
                "checked" if self.call_response.active == RESPONSE_JUNK else "",
            )
            .replace(
                "FUNCTION_RESPONSE_SPLIT_CHECKED",
                "checked" if self.call_response.split > 0 else "",
            )
            .replace("FUNCTION_RESPONSE_SPLIT_DELAY", str(self.call_response.split))
            .replace(
                "FUNCTION_RESPONSE_CR_CHECKED",
                "checked" if self.call_response.change_rate > 0 else "",
            )
            .replace("FUNCTION_RESPONSE_CR_PCT", str(self.call_response.change_rate))
            .replace("FUNCTION_RESPONSE_DELAY", str(self.call_response.delay))
            .replace("FUNCTION_RESPONSE_JUNK", str(self.call_response.junk_len))
            .replace("<!--FUNCTION_ERROR-->", function_error)
            .replace(
                "FUNCTION_RESPONSE_CLEAR_AFTER", str(self.call_response.clear_after)
            )
            .replace("<!--FC_ROWS-->", call_rows)
            .replace("<!--FC_FOOT-->", foot)
        )
        return new_html

    def build_html_log(self, _params, html):
        """Build html log page."""
        return html

    def build_html_server(self, _params, html):
        """Build html server page."""
        return html

    def build_json_registers(self, params):
        """Build json registers response."""
        # Process params using the helper function
        result_txt, foot = self.helper_handle_submit(params, {
            "Set": self.action_set,
        })

        if not result_txt:
            result_txt = "ok"
        if not foot:
            foot = "Operation completed successfully"

        # Extract necessary parameters
        try:
            range_start = int(params.get("range_start", 0))
            range_stop = int(params.get("range_stop", range_start))
        except ValueError:
            return {"result": "error", "error": "Invalid range parameters"}

        # Retrieve register details
        register_rows = []
        for i in range(range_start, range_stop + 1):
            inx, reg = self.datastore_context.get_text_register(i)
            row = {
                "index": inx,
                "type": reg.type,
                "access": reg.access,
                "action": reg.action,
                "value": reg.value,
                "count_read": reg.count_read,
                "count_write": reg.count_write
            }
            register_rows.append(row)

        # Generate register types and actions (assume these are predefined mappings)
        register_types = dict(self.datastore_context.registerType_name_to_id)
        register_actions = dict(self.datastore_context.action_name_to_id)

        # Build the JSON response
        json_response = {
            "result": result_txt,
            "footer": foot,
            "register_types": register_types,
            "register_actions": register_actions,
            "register_rows": register_rows,
        }

        return json_response

    def build_json_calls(self, params: dict) -> dict:
        """Build json calls response."""
        result_txt, foot = self.helper_handle_submit(params, {
            "Reset": self.action_reset,
            "Add": self.action_add,
            "Simulate": self.action_simulate,
        })
        if not foot:
            foot = "Monitoring active" if self.call_monitor.active else "not active"
        if not result_txt:
            result_txt = "ok"

        function_error = []
        for i, txt in (
            (1, "ILLEGAL_FUNCTION"),
            (2, "ILLEGAL_ADDRESS"),
            (3, "ILLEGAL_VALUE"),
            (4, "SLAVE_FAILURE"),
            (5, "ACKNOWLEDGE"),
            (6, "SLAVE_BUSY"),
            (7, "MEMORY_PARITY_ERROR"),
            (10, "GATEWAY_PATH_UNAVIABLE"),
            (11, "GATEWAY_NO_RESPONSE"),
        ):
            function_error.append({
                "value": i,
                "text": txt,
                "selected": i == self.call_response.error_response
            })

        range_start = (
            self.call_monitor.range_start
            if self.call_monitor.range_start != -1
            else None
        )
        range_stop = (
            self.call_monitor.range_stop
            if self.call_monitor.range_stop != -1
            else None
        )

        function_codes = []
        for function in self.request_lookup.values():
            function_codes.append({
                "value": function.function_code,
                "text": "function code name",
                "selected": function.function_code == self.call_monitor.function
            })

        simulation_action = "ACTIVE" if self.call_response.active != RESPONSE_INACTIVE else ""

        max_len = MAX_FILTER if self.call_monitor.active else 0
        while len(self.call_list) > max_len:
            del self.call_list[0]
        call_rows = []
        for entry in reversed(self.call_list):
            call_rows.append({
                "call": entry.call,
                "fc": entry.fc,
                "address": entry.address,
                "count": entry.count,
                "data": entry.data.decode()
            })

        json_response = {
            "simulation_action": simulation_action,
            "range_start": range_start,
            "range_stop": range_stop,
            "function_codes": function_codes,
            "function_show_hex_checked": self.call_monitor.hex,
            "function_show_decoded_checked": self.call_monitor.decode,
            "function_response_normal_checked": self.call_response.active == RESPONSE_NORMAL,
            "function_response_error_checked": self.call_response.active == RESPONSE_ERROR,
            "function_response_empty_checked": self.call_response.active == RESPONSE_EMPTY,
            "function_response_junk_checked": self.call_response.active == RESPONSE_JUNK,
            "function_response_split_checked": self.call_response.split > 0,
            "function_response_split_delay": self.call_response.split,
            "function_response_cr_checked": self.call_response.change_rate > 0,
            "function_response_cr_pct": self.call_response.change_rate,
            "function_response_delay": self.call_response.delay,
            "function_response_junk": self.call_response.junk_len,
            "function_error": function_error,
            "function_response_clear_after": self.call_response.clear_after,
            "call_rows": call_rows,
            "foot": foot,
            "result": result_txt
        }

        return json_response

    def build_json_log(self, params):
        """Build json log page."""
        return {"result": "error", "error": "log endpoint not implemented", "params": params}

    def build_json_server(self, params):
        """Build html server page."""
        return {"result": "error", "error": "server endpoint not implemented", "params": params}

    def helper_handle_submit(self, params, submit_actions):
        """Build html register submit."""
        try:
            range_start = int(params.get("range_start", -1))
        except ValueError:
            range_start = -1
        try:
            range_stop = int(params.get("range_stop", range_start))
        except ValueError:
            range_stop = -1
        if (submit := params["submit"]) not in submit_actions:
            return None, None
        return submit_actions[submit](params, range_start, range_stop)

    def action_clear(self, _params, _range_start, _range_stop):
        """Clear register filter."""
        self.register_filter = []
        return None, None

    def action_stop(self, _params, _range_start, _range_stop):
        """Stop call monitoring."""
        self.call_monitor = CallTypeMonitor()
        return None, "Stopped monitoring"

    def action_reset(self, _params, _range_start, _range_stop):
        """Reset call simulation."""
        self.call_response = CallTypeResponse()
        return None, None

    def action_add(self, params, range_start, range_stop):
        """Build list of registers matching filter."""
        reg_action = int(params.get("action", -1))
        reg_writeable = "writeable" in params
        reg_type = int(params.get("type", -1))
        filter_updated = 0
        if range_start != -1:
            steps = range(range_start, range_stop + 1)
        else:
            steps = range(1, self.datastore_context.register_count)
        for i in steps:
            if range_start != -1 and (i < range_start or i > range_stop):
                continue
            reg = self.datastore_context.registers[i]
            skip_filter = reg_writeable and not reg.access
            skip_filter |= reg_type not in (-1, reg.type)
            skip_filter |= reg_action not in (-1, reg.action)
            skip_filter |= i in self.register_filter
            if skip_filter:
                continue
            self.register_filter.append(i)
            filter_updated += 1
            if len(self.register_filter) >= MAX_FILTER:
                self.register_filter.sort()
                return None, f"Max. filter size {MAX_FILTER} exceeded!"
        self.register_filter.sort()
        return None, None

    def action_monitor(self, params, range_start, range_stop):
        """Start monitoring calls."""
        self.call_monitor.range_start = range_start
        self.call_monitor.range_stop = range_stop
        self.call_monitor.function = (
            int(params["function"]) if params["function"] else -1
        )
        self.call_monitor.hex = "show_hex" in params
        self.call_monitor.decode = "show_decode" in params
        self.call_monitor.active = True
        return None, None

    def action_set(self, params, _range_start, _range_stop):
        """Set register value."""
        if not (register := params["register"]):
            return "Missing register", None
        register = int(register)
        if value := params["value"]:
            self.datastore_context.registers[register].value = int(value)
        if bool(params.get("writeable", False)):
            self.datastore_context.registers[register].access = True
        return None, None

    def action_simulate(self, params, _range_start, _range_stop):
        """Simulate responses."""
        self.call_response.active = int(params["response_type"])
        if "response_split" in params:
            if params["split_delay"]:
                self.call_response.split = int(params["split_delay"])
            else:
                self.call_response.split = 1
        else:
            self.call_response.split = 0
        if "response_cr" in params:
            if params["response_cr_pct"]:
                self.call_response.change_rate = int(params["response_cr_pct"])
            else:
                self.call_response.change_rate = 0
        else:
            self.call_response.change_rate = 0
        if params["response_delay"]:
            self.call_response.delay = int(params["response_delay"])
        else:
            self.call_response.delay = 0
        if params["response_junk_datalen"]:
            self.call_response.junk_len = int(params["response_junk_datalen"])
        else:
            self.call_response.junk_len = 0
        self.call_response.error_response = int(params["response_error"])
        if params["response_clear_after"]:
            self.call_response.clear_after = int(params["response_clear_after"])
        else:
            self.call_response.clear_after = 1
        return None, None



# -*- coding: utf-8 -*-

"""package scinstr.vna
author    Benoit Dubois
copyright FEMTO ENGINEERING
license   GPL v3.0+
brief     Handle PNA-X, N522xA or N523x device from Keysight.
          Tested on N5234 device.
"""

import time
import socket
import struct
import logging
import numpy as np

PORT = 5025

#==============================================================================
class N523x(object):
    """Handle PNA-X, N522xA or N523x device from Keysight.
    """

    IFBW = (1, 2, 3, 5, 7, 10, 15, 20, 30, 50, 70, 100, 150, 200, 300, 500, 700,
            1e3, 1.5e3, 2e3, 3e3, 5e3, 7e3, 10e3, 15e3, 20e3, 30e3, 50e3, 70e3,
            100e3, 150e3, 200e3, 280e3, 360e3, 600e3)

    def __init__(self, ip, port=PORT, timeout=1.0):
        """Constructor.
        :param ip: IP address of device (str)
        :param port: Ethernet port of device (int)
        :param timeout: Timeout on connection instance (float)
        :returns: None
        """
        self.ip = ip
        self.port = port
        self._timeout = timeout
        self._sock = None

    def connect(self):
        """Connect to device.
        """
        try:
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._sock.settimeout(self._timeout)
            self._sock.connect((self.ip, self.port))
        except ValueError as ex:
            logging.error("Connection parameters out of range: %s", str(ex))
            return False
        except socket.timeout:
            logging.error("Timeout on connection")
            return False
        except Exception as ex:
            logging.error("Unexpected exception during connection with " + \
                          "VNA: %s", str(ex))
            return False
        else:
            logging.debug("Connected to VNA")
            return True

    def close(self):
        self._sock.close()

    def write(self, data):
        """"Ethernet writing process.
        :param data: data writes to device (str)
        :returns: None
        """
        try:
            self._sock.send((data + "\n").encode('utf-8'))
        except socket.timeout:
            logging.error("Timeout")
        except Exception as ex:
            logging.error(str(ex))
        logging.debug("write " + data.strip('\n'))

    def read(self, length=100):
        """Read process.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        try:
            retval = self._sock.recv(length).decode('utf-8').strip('\n')
        except socket.timeout:
            logging.error("Timeout")
            return ''
        except Exception as ex:
            logging.error(str(ex))
            raise ex
        logging.debug("read: " + retval)
        return retval

    def raw_read(self, length=512):
        """Raw read process.
        :param length: length of message to read (int)
        :returns: Message reads from device (bytes)
        """
        try:
            data = self._sock.recv(length)
        except socket.timeout:
            logging.error("Timeout")
            return bytes()
        except Exception as ex:
            logging.error(str(ex))
            raise ex
        return data

    def bin_read(self):
        """Read binary data then decode them to ascii.
        The reading process is specific to the transfert of binary data
        with these VNA devices: <header><data><EOT>, with:
        - <header>: #|lenght of bytes_count (one byte)|bytes_count
        - <data>: "REAL,64" (float64) binary data
        - <EOT>: '\n' character.
        Note: The data transfert format must be selected to "REAL,64"" before
        using this method
        """
        header_max_length = 11
        raw_data = self.raw_read(header_max_length)
        if raw_data.find(b'#') != 0:
            logging.error("Data header not valid")
            return
        byte_count_nb = int(raw_data[1:2])
        byte_count = int(raw_data[2:2+byte_count_nb])
        # Note : Read 'byte_count' bytes but only
        # 2 + byte_count_nb + byte_count - header_max_length
        # needs to be readen.
        # This tip can be used because EOF ('\n') is transmited at the end of
        # the message and thus stop reception of data.
        while len(raw_data) < byte_count:
            raw_data += self.raw_read(byte_count)
        nb_elem = int(byte_count / 8)
        data = np.asarray(struct.unpack(">{:d}d".format(nb_elem),
                                        raw_data[2+byte_count_nb:-1]))
        return data

    def query(self, msg, length=100):
        """Basic query process: write query then read response.
        """
        self.write(msg)
        return self.read(length)

    def reset(self):
        """Reset device.
        """
        self.write("*RST")

    @property
    def idn(self):
        """Return ID of device.
        """
        return self.query("*IDN?")

    def get_span(self, cnum=1):
        return self.query("SENS{}:FOM:RANG:SEGM:FREQ:SPAN?".format(cnum))

    def set_span(self, value, cnum=1):
        self.write("SENS{}:FREQ:SPAN {}".format(cnum, value))

    def get_start(self, cnum=1):
        return self.query("SENS{}:FREQ:START?".format(cnum))

    def set_start(self, value, cnum=1):
        self.write("SENS{}:FREQ:START {}".format(cnum, value))

    def get_stop(self, cnum=1):
        return self.query("SENS{}:FREQ:STOP?".format(cnum))

    def set_stop(self, value, cnum=1):
        self.write("SENS{}:FREQ:STOP {}".format(cnum, value))

    def get_center_freq(self, cnum=1):
        return self.query("SENS{}:FREQ:CENT?".format(cnum))

    def set_center_freq(self, value, cnum=1):
        self.write("SENS{}:FREQ:CENT {}".format(cnum, value))

    def get_points(self, cnum=1):
        return self.query("SENS{}:SWE:POINTS?".format(cnum))

    def set_points(self, value, cnum=1):
        self.write("SENS{}:SWE:POINTS {}".format(cnum, value))

    def get_sweep_type(self, cnum=1):
        return self.query("SENS{}:SWE:TYPE?".format(cnum))

    def set_sweep_type(self, value, cnum=1):
        self.write("SENS{}:SWE:TYPE {}".format(cnum, value))

    @staticmethod
    def read_s2p(filename):
        return np.loadtxt(filename, comments=['!', '#'])

    def get_window_numbers(self):
        """Return the number of existing windows.
        """
        data = self.query("DISP:CAT?")
        if data is None:
            return []
        return [int(x) for x in data.replace('\"', '').split(',')]

    def get_measurement_catalog(self, channel=''):
        """Returns ALL measurement numbers, or measurement numbers
        from a specified channel
        :param channel: Channel number to catalog. If not specified,
                        all measurement numbers are returned.
        :returns: ALL measurement numbers, or measurement numbers
                  from a specified channel
        """
        data = self.query("SYST:MEAS:CAT? {}".format(channel))
        if data is None:
            return []
        return [int(x) for x in data.replace('\"', '').split(',')]

    def measurement_number_to_trace(self, nb=None):
        """Returns the trace number of the specified measurement number.
        Trace numbers restart for each window while measurement numbers are
        always unique.
        :param n: Measurement number for which to return the trace number.
                  If unspecified, value is set to 1.
        """
        return self.query("SYST:MEAS{}:TRAC?".format(nb))

    def measurement_number_to_name(self, nb=None):
        """Returns the name of the specified measurement.
        :param n: Measurement number for which to return the measurement name.
        If unspecified, value is set to 1.
        """
        return self.query("SYST:MEAS{}:NAME?".format(nb)).replace('\"', '')

    def set_measurement(self, name, fast=True):
        """ Sets the selected measurement. Most CALC: commands require that
        this command be sent before a setting change is made. One measurement
        on each channel can be selected at the same time.
        :param name: Name of the measurement. CASE-SENSITIVE. Do NOT include
                      the parameter name that is returned with Calc:Par:Cat?
        :param fast: Optional. The PNA display is NOT updated. Therefore,
                     do not use this argument when an operator is using the
                     PNA display. Otherwise, sending this argument results
                     in much faster sweep speeds. There is NO other reason
                     to NOT send this argument.
        """
        if name is None:
            logging.error("Requiered name parameter")
            raise ValueError("Requiered name parameter")
        cnum = int(name[2])
        self.write("CALC{}:PAR:SEL {}{}"
                   .format(cnum, name, ",fast" if fast is True else None))

    def get_measurement(self, name):
        """Get a data measurement.
        Note that VNA must be configured to transfer data in float 64 format
        before using this method.
        :param name: Name of the measurement. CASE-SENSITIVE. Do NOT include
                     the parameter name that is returned with Calc:Par:Cat?
        :return: Array of measurement data.
        """
        cnum = int(name[2])
        self.set_measurement(name)
        if self.get_sweep_type() != "LIN":
            raise NotImplementedError
        datax = np.linspace(float(self.get_start(cnum)),
                            float(self.get_stop(cnum)),
                            int(self.get_points(cnum)))
        self.write("CALC{}:DATA? FDATA".format(cnum))
        datay = self.bin_read()
        retval = np.asarray([datax, datay])
        return retval

    def get_snp(self, cnum=1):
        """Get snp data.
        :param cnum: Channel number of the measurement. There must be a selected
        measurement on that channel. If unspecified, <cnum> is set to 1.
        :return: Array of data.
        """
        self.write("CALC{}:DATA:SNP:PORT? \"1,2\"".format(cnum))
        return np.asarray(self.bin_read()).reshape(9, -1)

    def get_measurements(self):
        """Find current measurements, get data then prepare a list of array
        [freq, data] for each measurement.
        :returns: list of measurements
        """
        datas = []
        meas_nb = self.get_measurement_catalog()
        for nb in meas_nb:
            name = self.measurement_number_to_name(nb)
            datas.append(self.get_measurement(name))
        return datas

    def get_snps(self):
        """Get all snp traces.
        """
        meas_nb = self.get_measurement_catalog()
        datas = []
        for measurement in meas_nb:
            datas.append(self.get_snp(measurement))
        return datas

    def write_snps(self, filename=None):
        """Write all snp traces.
        """
        if filename is None:
            filename = time.strftime("%Y%m%d-%H%M%S", time.gmtime())
        meas_nb = self.get_measurement_catalog()
        for measurement in meas_nb:
            data = np.asarray(self.get_snp(measurement)).reshape(9, -1)
            np.savetxt(filename + '_' + str(measurement) + '.s2p',
                       data,
                       delimiter='\t',
                       header="#f(Hz)\tReal(S11)\tImag(S11)" +
                       "\tReal(S21)\tImag(S21)" +
                       "\tReal(S12)\tImag(S12)" +
                       "\tReal(S22)\tImag(S22)")


#==============================================================================
def main():
    """Main of test program
    :returns: None
    """
    import matplotlib.pyplot as plt

    ip = "192.168.0.11"

    acp_type = 'data' # 'data' or 'snp'

    dev = N523x(ip, timeout=0.5)

    if dev.connect() is False:
        print("Connection error")
        return

    print("Connected to", dev.id)

    dev.write("FORMAT:DATA REAL,64")

    if acp_type == 'snp':
        dev.write("MMEM:STOR:TRAC:FORM:SNP MA")
        datas = dev.get_snp()
        plt.figure(1)
        nb_rows = datas.shape[0]
        for idx, data in enumerate(datas):
            plt.subplot(nb_rows*100+10+idx)
            plt.plot(datas[0], data, 'b')
        plt.show()

    else:
        datas = dev.get_measurements()
        plt.figure(1)
        nb_rows = len(datas)
        for idx, data in enumerate(datas):
            plt.subplot(nb_rows*100+10+idx+1)
            plt.ylabel(dev.measurement_number_to_name(idx+1))
            try:
                plt.plot(data[0], data[1], 'b', )
            except Exception:
                pass
        plt.show()


#==============================================================================
if __name__ == '__main__':
    CONSOLE_LOG_LEVEL = logging.DEBUG
    CONSOLE_FORMAT = '%(levelname) -8s %(filename)s (%(lineno)d): %(message)s'
    logging.basicConfig(format=CONSOLE_FORMAT, level=CONSOLE_LOG_LEVEL)

    main()



"""Pymodbus ModbusSimulatorContext."""
from __future__ import annotations

import dataclasses
import random
import struct
from collections.abc import Callable
from datetime import datetime
from typing import Any

from pymodbus.datastore.context import ModbusBaseSlaveContext


WORD_SIZE = 16


@dataclasses.dataclass(frozen=True)
class CellType:
    """Define single cell types."""

    INVALID: int = 0
    BITS: int = 1
    UINT16: int = 2
    UINT32: int = 3
    FLOAT32: int = 4
    STRING: int = 5
    NEXT: int = 6


@dataclasses.dataclass(repr=False, eq=False)
class Cell:
    """Handle a single cell."""

    type: int = CellType.INVALID
    access: bool = False
    value: int = 0
    action: int = 0
    action_parameters: dict[str, Any] | None = None
    count_read: int = 0
    count_write: int = 0


class TextCell:  # pylint: disable=too-few-public-methods
    """A textual representation of a single cell."""

    type: str
    access: str
    value: str
    action: str
    action_parameters: str
    count_read: str
    count_write: str


@dataclasses.dataclass
class Label:  # pylint: disable=too-many-instance-attributes
    """Defines all dict values.

    :meta private:
    """

    action: str = "action"
    addr: str = "addr"
    any: str = "any"
    co_size: str = "co size"
    defaults: str = "defaults"
    di_size: str = "di size"
    hr_size: str = "hr size"
    increment: str = "increment"
    invalid: str = "invalid"
    ir_size: str = "ir size"
    parameters: str = "parameters"
    method: str = "method"
    next: str = "next"
    none: str = "none"
    random: str = "random"
    repeat: str = "repeat"
    reset: str = "reset"
    setup: str = "setup"
    shared_blocks: str = "shared blocks"
    timestamp: str = "timestamp"
    repeat_to: str = "to"
    type: str = "type"
    type_bits = "bits"
    type_exception: str = "type exception"
    type_uint16: str = "uint16"
    type_uint32: str = "uint32"
    type_float32: str = "float32"
    type_string: str = "string"
    uptime: str = "uptime"
    value: str = "value"
    write: str = "write"

    @classmethod
    def try_get(cls, key, config_part):
        """Check if entry is present in config."""
        if key not in config_part:
            txt = f"ERROR Configuration invalid, missing {key} in {config_part}"
            raise RuntimeError(txt)
        return config_part[key]


class Setup:
    """Setup simulator.

    :meta private:
    """

    def __init__(self, runtime: Any) -> None:
        """Initialize."""
        self.runtime = runtime
        self.config: Any = {}
        self.config_types: dict[str, dict[str, Any]] = {
            Label.type_bits: {
                Label.type: CellType.BITS,
                Label.next: None,
                Label.value: 0,
                Label.action: None,
                Label.method: self.handle_type_bits,
            },
            Label.type_uint16: {
                Label.type: CellType.UINT16,
                Label.next: None,
                Label.value: 0,
                Label.action: None,
                Label.method: self.handle_type_uint16,
            },
            Label.type_uint32: {
                Label.type: CellType.UINT32,
                Label.next: CellType.NEXT,
                Label.value: 0,
                Label.action: None,
                Label.method: self.handle_type_uint32,
            },
            Label.type_float32: {
                Label.type: CellType.FLOAT32,
                Label.next: CellType.NEXT,
                Label.value: 0,
                Label.action: None,
                Label.method: self.handle_type_float32,
            },
            Label.type_string: {
                Label.type: CellType.STRING,
                Label.next: CellType.NEXT,
                Label.value: 0,
                Label.action: None,
                Label.method: self.handle_type_string,
            },
        }

    def handle_type_bits(self, start, stop, value, action, action_parameters):
        """Handle type bits."""
        for reg in self.runtime.registers[start:stop]:
            if reg.type != CellType.INVALID:
                raise RuntimeError(f'ERROR "{Label.type_bits}" {reg} used')
            reg.value = value
            reg.type = CellType.BITS
            reg.action = action
            reg.action_parameters = action_parameters

    def handle_type_uint16(self, start, stop, value, action, action_parameters):
        """Handle type uint16."""
        for reg in self.runtime.registers[start:stop]:
            if reg.type != CellType.INVALID:
                raise RuntimeError(f'ERROR "{Label.type_uint16}" {reg} used')
            reg.value = value
            reg.type = CellType.UINT16
            reg.action = action
            reg.action_parameters = action_parameters

    def handle_type_uint32(self, start, stop, value, action, action_parameters):
        """Handle type uint32."""
        regs_value = ModbusSimulatorContext.build_registers_from_value(value, True)
        for i in range(start, stop, 2):
            regs = self.runtime.registers[i : i + 2]
            if regs[0].type != CellType.INVALID or regs[1].type != CellType.INVALID:
                raise RuntimeError(f'ERROR "{Label.type_uint32}" {i},{i + 1} used')
            regs[0].value = regs_value[0]
            regs[0].type = CellType.UINT32
            regs[0].action = action
            regs[0].action_parameters = action_parameters
            regs[1].value = regs_value[1]
            regs[1].type = CellType.NEXT

    def handle_type_float32(self, start, stop, value, action, action_parameters):
        """Handle type uint32."""
        regs_value = ModbusSimulatorContext.build_registers_from_value(value, False)
        for i in range(start, stop, 2):
            regs = self.runtime.registers[i : i + 2]
            if regs[0].type != CellType.INVALID or regs[1].type != CellType.INVALID:
                raise RuntimeError(f'ERROR "{Label.type_float32}" {i},{i + 1} used')
            regs[0].value = regs_value[0]
            regs[0].type = CellType.FLOAT32
            regs[0].action = action
            regs[0].action_parameters = action_parameters
            regs[1].value = regs_value[1]
            regs[1].type = CellType.NEXT

    def handle_type_string(self, start, stop, value, action, action_parameters):
        """Handle type string."""
        regs = stop - start
        reg_len = regs * 2
        if len(value) > reg_len:
            raise RuntimeError(
                f'ERROR "{Label.type_string}" {start} too long "{value}"'
            )
        value = value.ljust(reg_len)
        for i in range(stop - start):
            reg = self.runtime.registers[start + i]
            if reg.type != CellType.INVALID:
                raise RuntimeError(f'ERROR "{Label.type_string}" {start + i} used')
            j = i * 2
            reg.value = int.from_bytes(bytes(value[j : j + 2], "UTF-8"), "big")
            reg.type = CellType.NEXT
        self.runtime.registers[start].type = CellType.STRING
        self.runtime.registers[start].action = action
        self.runtime.registers[start].action_parameters = action_parameters

    def handle_setup_section(self):
        """Load setup section."""
        layout = Label.try_get(Label.setup, self.config)
        self.runtime.fc_offset = dict.fromkeys(range(25), 0)
        size_co = Label.try_get(Label.co_size, layout)
        size_di = Label.try_get(Label.di_size, layout)
        size_hr = Label.try_get(Label.hr_size, layout)
        size_ir = Label.try_get(Label.ir_size, layout)
        if Label.try_get(Label.shared_blocks, layout):
            total_size = max(size_co, size_di, size_hr, size_ir)
        else:
            # set offset (block) for each function code
            # starting with fc = 1, 5, 15
            self.runtime.fc_offset[2] = size_co
            total_size = size_co + size_di
            self.runtime.fc_offset[4] = total_size
            total_size += size_ir
            for i in (3, 6, 16, 22, 23):
                self.runtime.fc_offset[i] = total_size
            total_size += size_hr
        first_cell = Cell()
        self.runtime.registers = [
            dataclasses.replace(first_cell) for i in range(total_size)
        ]
        self.runtime.register_count = total_size
        self.runtime.type_exception = bool(Label.try_get(Label.type_exception, layout))
        defaults = Label.try_get(Label.defaults, layout)
        defaults_value = Label.try_get(Label.value, defaults)
        defaults_action = Label.try_get(Label.action, defaults)
        for key, entry in self.config_types.items():
            entry[Label.value] = Label.try_get(key, defaults_value)
            if (
                action := Label.try_get(key, defaults_action)
            ) not in self.runtime.action_name_to_id:
                raise RuntimeError(f"ERROR illegal action {key} in {defaults_action}")
            entry[Label.action] = action
        del self.config[Label.setup]

    def handle_invalid_address(self):
        """Handle invalid address."""
        for entry in Label.try_get(Label.invalid, self.config):
            if isinstance(entry, int):
                entry = [entry, entry]
            for i in range(entry[0], entry[1] + 1):
                if i >= self.runtime.register_count:
                    raise RuntimeError(
                        f'Error section "{Label.invalid}" addr {entry} out of range'
                    )
                reg = self.runtime.registers[i]
                reg.type = CellType.INVALID
        del self.config[Label.invalid]

    def handle_write_allowed(self):
        """Handle write allowed."""
        for entry in Label.try_get(Label.write, self.config):
            if isinstance(entry, int):
                entry = [entry, entry]
            for i in range(entry[0], entry[1] + 1):
                if i >= self.runtime.register_count:
                    raise RuntimeError(
                        f'Error section "{Label.write}" addr {entry} out of range'
                    )
                reg = self.runtime.registers[i]
                if reg.type == CellType.INVALID:
                    txt = f'ERROR Configuration invalid in section "write" register {i} not defined'
                    raise RuntimeError(txt)
                reg.access = True
        del self.config[Label.write]

    def handle_types(self):
        """Handle the different types."""
        for section, type_entry in self.config_types.items():
            layout = Label.try_get(section, self.config)
            for entry in layout:
                if not isinstance(entry, dict):
                    entry = {Label.addr: entry}
                regs = Label.try_get(Label.addr, entry)
                if not isinstance(regs, list):
                    regs = [regs, regs]
                start = regs[0]
                if (stop := regs[1]) >= self.runtime.register_count:
                    raise RuntimeError(f'Error "{section}" {start}, {stop} illegal')
                type_entry[Label.method](
                    start,
                    stop + 1,
                    entry.get(Label.value, type_entry[Label.value]),
                    self.runtime.action_name_to_id[
                        entry.get(Label.action, type_entry[Label.action])
                    ],
                    entry.get(Label.parameters, None),
                )
            del self.config[section]

    def handle_repeat(self):
        """Handle repeat."""
        for entry in Label.try_get(Label.repeat, self.config):
            addr = Label.try_get(Label.addr, entry)
            copy_start = addr[0]
            copy_end = addr[1]
            copy_inx = copy_start - 1
            addr_to = Label.try_get(Label.repeat_to, entry)
            for inx in range(addr_to[0], addr_to[1] + 1):
                copy_inx = copy_start if copy_inx >= copy_end else copy_inx + 1
                if inx >= self.runtime.register_count:
                    raise RuntimeError(
                        f'Error section "{Label.repeat}" entry {entry} out of range'
                    )
                self.runtime.registers[inx] = dataclasses.replace(
                    self.runtime.registers[copy_inx]
                )
        del self.config[Label.repeat]

    def setup(self, config, custom_actions) -> None:
        """Load layout from dict with json structure."""
        actions = {
            Label.increment: self.runtime.action_increment,
            Label.random: self.runtime.action_random,
            Label.reset: self.runtime.action_reset,
            Label.timestamp: self.runtime.action_timestamp,
            Label.uptime: self.runtime.action_uptime,
        }
        if custom_actions:
            actions.update(custom_actions)
        self.runtime.action_name_to_id = {None: 0}
        self.runtime.action_id_to_name = [Label.none]
        self.runtime.action_methods = [None]
        i = 1
        for key, method in actions.items():
            self.runtime.action_name_to_id[key] = i
            self.runtime.action_id_to_name.append(key)
            self.runtime.action_methods.append(method)
            i += 1
        self.runtime.registerType_name_to_id = {
            Label.type_bits: CellType.BITS,
            Label.type_uint16: CellType.UINT16,
            Label.type_uint32: CellType.UINT32,
            Label.type_float32: CellType.FLOAT32,
            Label.type_string: CellType.STRING,
            Label.next: CellType.NEXT,
            Label.invalid: CellType.INVALID,
        }
        self.runtime.registerType_id_to_name = [None] * len(
            self.runtime.registerType_name_to_id
        )
        for name, cell_type in self.runtime.registerType_name_to_id.items():
            self.runtime.registerType_id_to_name[cell_type] = name

        self.config = config
        self.handle_setup_section()
        self.handle_invalid_address()
        self.handle_types()
        self.handle_write_allowed()
        self.handle_repeat()
        if self.config:
            raise RuntimeError(f"INVALID key in setup: {self.config}")


class ModbusSimulatorContext(ModbusBaseSlaveContext):
    """Modbus simulator.

    :param config: A dict with structure as shown below.
    :param actions: A dict with "<name>": <function> structure.
    :raises RuntimeError: if json contains errors (msg explains what)

    It builds and maintains a virtual copy of a device, with simulation of
    device specific functions.

    The device is described in a dict, user supplied actions will
    be added to the builtin actions.

    It is used in conjunction with a pymodbus server.

    Example::

        store = ModbusSimulatorContext(<config dict>, <actions dict>)
        StartAsyncTcpServer(<host>, context=store)

        Now the server will simulate the defined device with features like:

        - invalid addresses
        - write protected addresses
        - optional control of access for string, uint32, bit/bits
        - builtin actions for e.g. reset/datetime, value increment by read
        - custom actions

    Description of the json file or dict to be supplied::

        {
            "setup": {
                "di size": 0,  --> Size of discrete input block (8 bit)
                "co size": 0,  --> Size of coils block (8 bit)
                "ir size": 0,  --> Size of input registers block (16 bit)
                "hr size": 0,  --> Size of holding registers block (16 bit)
                "shared blocks": True,  --> share memory for all blocks (largest size wins)
                "defaults": {
                    "value": {  --> Initial values (can be overwritten)
                        "bits": 0x01,
                        "uint16": 122,
                        "uint32": 67000,
                        "float32": 127.4,
                        "string": " ",
                    },
                    "action": {  --> default action (can be overwritten)
                        "bits": None,
                        "uint16": None,
                        "uint32": None,
                        "float32": None,
                        "string": None,
                    },
                },
                "type exception": False,  --> return IO exception if read/write on non boundary
            },
            "invalid": [  --> List of invalid addresses, IO exception returned
                51,                --> single register
                [78, 99],         --> start, end registers, repeated as needed
            ],
            "write": [   --> allow write, efault is ReadOnly
                [5, 5]  --> start, end bytes, repeated as needed
            ],
            "bits": [  --> Define bits (1 register == 2 bytes)
                [30, 31],  --> start, end registers, repeated as needed
                {"addr": [32, 34], "value": 0xF1},  --> with value
                {"addr": [35, 36], "action": "increment"},  --> with action
                {"addr": [37, 38], "action": "increment", "value": 0xF1}  --> with action and value
                {"addr": [37, 38], "action": "increment", "parameters": {"min": 0, "max": 100}}  --> with action with arguments
            ],
            "uint16": [  --> Define uint16 (1 register == 2 bytes)
                --> same as type_bits
            ],
            "uint32": [  --> Define 32 bit integers (2 registers == 4 bytes)
                --> same as type_bits
            ],
            "float32": [  --> Define 32 bit floats (2 registers == 4 bytes)
                --> same as type_bits
            ],
            "string": [  --> Define strings (variable number of registers (each 2 bytes))
                [21, 22],  --> start, end registers, define 1 string
                {"addr": 23, 25], "value": "ups"},  --> with value
                {"addr": 26, 27], "action": "user"},  --> with action
                {"addr": 28, 29], "action": "", "value": "user"}  --> with action and value
            ],
            "repeat": [ --> allows to repeat section e.g. for n devices
                {"addr": [100, 200], "to": [50, 275]}   --> Repeat registers 100-200 to 50+ until 275
            ]
        }
    """

    # --------------------------------------------
    # External interfaces
    # --------------------------------------------
    start_time = int(datetime.now().timestamp())

    def __init__(
        self, config: dict[str, Any], custom_actions: dict[str, Callable] | None
    ) -> None:
        """Initialize."""
        self.registers: list[Cell] = []
        self.fc_offset: dict[int, int] = {}
        self.register_count = 0
        self.type_exception = False
        self.action_name_to_id: dict[str, int] = {}
        self.action_id_to_name: list[str] = []
        self.action_methods: list[Callable] = []
        self.registerType_name_to_id: dict[str, int] = {}
        self.registerType_id_to_name: list[str] = []
        Setup(self).setup(config, custom_actions)

    # --------------------------------------------
    # Simulator server interface
    # --------------------------------------------
    def get_text_register(self, register):
        """Get raw register."""
        reg = self.registers[register]
        text_cell = TextCell()
        text_cell.type = self.registerType_id_to_name[reg.type]
        text_cell.access = str(reg.access)
        text_cell.count_read = str(reg.count_read)
        text_cell.count_write = str(reg.count_write)
        text_cell.action = self.action_id_to_name[reg.action]
        if reg.action_parameters:
            text_cell.action = f"{text_cell.action}({reg.action_parameters})"
        if reg.type in (CellType.INVALID, CellType.UINT16, CellType.NEXT):
            text_cell.value = str(reg.value)
            build_len = 0
        elif reg.type == CellType.BITS:
            text_cell.value = hex(reg.value)
            build_len = 0
        elif reg.type == CellType.UINT32:
            tmp_regs = [reg.value, self.registers[register + 1].value]
            text_cell.value = str(self.build_value_from_registers(tmp_regs, True))
            build_len = 1
        elif reg.type == CellType.FLOAT32:
            tmp_regs = [reg.value, self.registers[register + 1].value]
            text_cell.value = str(self.build_value_from_registers(tmp_regs, False))
            build_len = 1
        else:  # reg.type == CellType.STRING:
            j = register
            text_cell.value = ""
            while True:
                text_cell.value += str(
                    self.registers[j].value.to_bytes(2, "big"),
                    encoding="utf-8",
                    errors="ignore",
                )
                j += 1
                if self.registers[j].type != CellType.NEXT:
                    break
            build_len = j - register - 1
        reg_txt = f"{register}-{register + build_len}" if build_len else f"{register}"
        return reg_txt, text_cell

    # --------------------------------------------
    # Modbus server interface
    # --------------------------------------------

    _write_func_code = (5, 6, 15, 16, 22, 23)
    _bits_func_code = (1, 2, 5, 15)

    def loop_validate(self, address, end_address, fx_write):
        """Validate entry in loop.

        :meta private:
        """
        i = address
        while i < end_address:
            reg = self.registers[i]
            if (fx_write and not reg.access) or reg.type == CellType.INVALID:
                return False
            if not self.type_exception:
                i += 1
                continue
            if reg.type == CellType.NEXT:
                return False
            if reg.type in (CellType.BITS, CellType.UINT16):
                i += 1
            elif reg.type in (CellType.UINT32, CellType.FLOAT32):
                if i + 1 >= end_address:
                    return False
                i += 2
            else:
                i += 1
                while i < end_address:
                    if self.registers[i].type == CellType.NEXT:
                        i += 1
        return True

    def validate(self, func_code, address, count=1):
        """Check to see if the request is in range.

        :meta private:
        """
        if func_code in self._bits_func_code:
            # Bit count, correct to register count
            count = int((count + WORD_SIZE - 1) / WORD_SIZE)
            address = int(address / 16)

        real_address = self.fc_offset[func_code] + address
        if real_address < 0 or real_address > self.register_count:
            return False

        fx_write = func_code in self._write_func_code
        return self.loop_validate(real_address, real_address + count, fx_write)

    def getValues(self, func_code, address, count=1):
        """Return the requested values of the datastore.

        :meta private:
        """
        if not self.validate(func_code, address, count):
            return 2
        result = []
        if func_code not in self._bits_func_code:
            real_address = self.fc_offset[func_code] + address
            for i in range(real_address, real_address + count):
                reg = self.registers[i]
                parameters = reg.action_parameters if reg.action_parameters else {}
                if reg.action:
                    self.action_methods[reg.action](self.registers, i, reg, **parameters)
                self.registers[i].count_read += 1
                result.append(reg.value)
        else:
            # bit access
            real_address = self.fc_offset[func_code] + int(address / 16)
            bit_index = address % 16
            reg_count = int((count + bit_index + 15) / 16)
            for i in range(real_address, real_address + reg_count):
                reg = self.registers[i]
                if reg.action:
                    parameters = reg.action_parameters or {}
                    self.action_methods[reg.action](
                        self.registers, i, reg, **parameters
                    )
                self.registers[i].count_read += 1
                while count and bit_index < 16:
                    result.append(bool(reg.value & (2**bit_index)))
                    count -= 1
                    bit_index += 1
                bit_index = 0
        return result

    def setValues(self, func_code, address, values):
        """Set the requested values of the datastore.

        :meta private:
        """
        if func_code not in self._bits_func_code:
            real_address = self.fc_offset[func_code] + address
            for value in values:
                if not self.validate(func_code, address):
                    return 2
                self.registers[real_address].value = value
                self.registers[real_address].count_write += 1
                real_address += 1
                address += 1
            return None

        # bit access
        real_address = self.fc_offset[func_code] + int(address / 16)
        bit_index = address % 16
        for value in values:
            if not self.validate(func_code, address):
                return 2
            bit_mask = 2**bit_index
            if bool(value):
                self.registers[real_address].value |= bit_mask
            else:
                self.registers[real_address].value &= ~bit_mask
            self.registers[real_address].count_write += 1
            bit_index += 1
            if bit_index == 16:
                bit_index = 0
                real_address += 1
                address += 1
        return None

    # --------------------------------------------
    # Internal action methods
    # --------------------------------------------

    @classmethod
    def action_random(cls, registers, inx, cell, minval=1, maxval=65536):
        """Update with random value.

        :meta private:
        """
        if cell.type in (CellType.BITS, CellType.UINT16):
            registers[inx].value = random.randint(int(minval), int(maxval))
        elif cell.type == CellType.FLOAT32:
            regs = cls.build_registers_from_value(
                random.uniform(float(minval), float(maxval)), False
            )
            registers[inx].value = regs[0]
            registers[inx + 1].value = regs[1]
        elif cell.type == CellType.UINT32:
            regs = cls.build_registers_from_value(
                random.randint(int(minval), int(maxval)), True
            )
            registers[inx].value = regs[0]
            registers[inx + 1].value = regs[1]

    @classmethod
    def action_increment(cls, registers, inx, cell, minval=None, maxval=None):
        """Increment value reset with overflow.

        :meta private:
        """
        reg = registers[inx]
        reg2 = registers[inx + 1]
        if cell.type in (CellType.BITS, CellType.UINT16):
            value = reg.value + 1
            if maxval and value > maxval:
                value = minval
            if minval and value < minval:
                value = minval
            reg.value = value
        elif cell.type == CellType.FLOAT32:
            tmp_reg = [reg.value, reg2.value]
            value = cls.build_value_from_registers(tmp_reg, False)
            value += 1.0
            if maxval and value > maxval:
                value = minval
            if minval and value < minval:
                value = minval
            new_regs = cls.build_registers_from_value(value, False)
            reg.value = new_regs[0]
            reg2.value = new_regs[1]
        elif cell.type == CellType.UINT32:
            tmp_reg = [reg.value, reg2.value]
            value = cls.build_value_from_registers(tmp_reg, True)
            value += 1
            if maxval and value > maxval:
                value = minval
            if minval and value < minval:
                value = minval
            new_regs = cls.build_registers_from_value(value, True)
            reg.value = new_regs[0]
            reg2.value = new_regs[1]

    @classmethod
    def action_timestamp(cls, registers, inx, _cell, **_parameters):
        """Set current time.

        :meta private:
        """
        system_time = datetime.now()
        registers[inx].value = system_time.year
        registers[inx + 1].value = system_time.month - 1
        registers[inx + 2].value = system_time.day
        registers[inx + 3].value = system_time.weekday() + 1
        registers[inx + 4].value = system_time.hour
        registers[inx + 5].value = system_time.minute
        registers[inx + 6].value = system_time.second

    @classmethod
    def action_reset(cls, _registers, _inx, _cell, **_parameters):
        """Reboot server.

        :meta private:
        """
        raise RuntimeError("RESET server")

    @classmethod
    def action_uptime(cls, registers, inx, cell, **_parameters):
        """Return uptime in seconds.

        :meta private:
        """
        value = int(datetime.now().timestamp()) - cls.start_time + 1

        if cell.type in (CellType.BITS, CellType.UINT16):
            registers[inx].value = value
        elif cell.type == CellType.FLOAT32:
            regs = cls.build_registers_from_value(value, False)
            registers[inx].value = regs[0]
            registers[inx + 1].value = regs[1]
        elif cell.type == CellType.UINT32:
            regs = cls.build_registers_from_value(value, True)
            registers[inx].value = regs[0]
            registers[inx + 1].value = regs[1]

    # --------------------------------------------
    # Internal helper methods
    # --------------------------------------------

    @classmethod
    def build_registers_from_value(cls, value, is_int):
        """Build registers from int32 or float32."""
        regs = [0, 0]
        if is_int:
            value_bytes = int.to_bytes(value, 4, "big")
        else:
            value_bytes = struct.pack(">f", value)
        regs[0] = int.from_bytes(value_bytes[:2], "big")
        regs[1] = int.from_bytes(value_bytes[-2:], "big")
        return regs

    @classmethod
    def build_value_from_registers(cls, registers, is_int):
        """Build int32 or float32 value from registers."""
        value_bytes = int.to_bytes(registers[0], 2, "big") + int.to_bytes(
            registers[1], 2, "big"
        )
        if is_int:
            value = int.from_bytes(value_bytes, "big")
        else:
            value = struct.unpack(">f", value_bytes)[0]
        return value



"""Collection of transaction based abstractions."""
from __future__ import annotations

import asyncio
from collections.abc import Callable
from threading import RLock

from pymodbus.exceptions import ConnectionException, ModbusIOException
from pymodbus.framer import FramerAscii, FramerBase, FramerRTU
from pymodbus.logging import Log
from pymodbus.pdu import ExceptionResponse, ModbusPDU
from pymodbus.transport import CommParams, ModbusProtocol


class TransactionManager(ModbusProtocol):
    """Transaction manager.

    This is the central class of the library, providing a separation between API and communication:
    - clients/servers calls the manager to execute requests/responses
    - transport/framer/pdu is by the manager to communicate with the devices

    Transaction manager handles:
    - Execution of requests (client), with retries and locking
    - Sending of responses (server), with retries
    - Connection management (on top of what transport offers)
    - No response (temporarily) from a device

    Transaction manager offers:
    - a simple execute interface for requests (client)
    - a simple send interface for responses (server)
    - external trace methods tracing outgoing/incoming packets/PDUs (byte stream)
    """

    def __init__(
        self,
        params: CommParams,
        framer: FramerBase,
        retries: int,
        is_server: bool,
        trace_packet: Callable[[bool, bytes], bytes] | None,
        trace_pdu: Callable[[bool, ModbusPDU], ModbusPDU] | None,
        trace_connect: Callable[[bool], None] | None,
        sync_client = None,
        ) -> None:
        """Initialize an instance of the ModbusTransactionManager."""
        self.is_sync = bool(sync_client)
        super().__init__(params, is_server, is_sync=self.is_sync)
        self.framer = framer
        self.retries = retries
        self.next_tid: int = 0
        self.request_dev_id: int = 0
        self.request_transaction_id: int = 0
        self.trace_packet = trace_packet or self.dummy_trace_packet
        self.trace_pdu = trace_pdu or self.dummy_trace_pdu
        self.trace_connect = trace_connect or self.dummy_trace_connect
        self.max_until_disconnect = self.count_until_disconnect = retries + 3
        if sync_client:
            self.sync_client = sync_client
            self._sync_lock = RLock()
            self.low_level_send = self.sync_client.send
        else:
            self._lock = asyncio.Lock()
            self.low_level_send = self.send
            self.response_future: asyncio.Future = asyncio.Future()
            self.last_pdu: ModbusPDU | None
            self.last_addr: tuple | None

    def dummy_trace_packet(self, sending: bool, data: bytes) -> bytes:
        """Do dummy trace."""
        _ = sending
        return data

    def dummy_trace_pdu(self, sending: bool, pdu: ModbusPDU) -> ModbusPDU:
        """Do dummy trace."""
        _ = sending
        return pdu

    def dummy_trace_connect(self, connect: bool) -> None:
        """Do dummy trace."""
        _ = connect

    def sync_get_response(self, dev_id) -> ModbusPDU:
        """Receive until PDU is correct or timeout."""
        databuffer = b''
        while True:
            if not (data := self.sync_client.recv(None)):
                raise asyncio.exceptions.TimeoutError()

            if self.sent_buffer:
                if data.startswith(self.sent_buffer):
                    Log.debug(
                        "sync recv skipping (local_echo): {}",
                        self.sent_buffer,
                        ":hex",
                    )
                    data = data[len(self.sent_buffer) :]
                    self.sent_buffer = b""
                elif self.sent_buffer.startswith(data):
                    Log.debug(
                        "sync recv skipping (partial local_echo): {}", data, ":hex"
                    )
                    self.sent_buffer = self.sent_buffer[len(data) :]
                    continue
                else:
                    Log.debug("did not sync receive local echo: {}", data, ":hex")
                    self.sent_buffer = b""
                if not data:
                    continue

            databuffer += data
            used_len, pdu = self.framer.processIncomingFrame(self.trace_packet(False, databuffer))
            databuffer = databuffer[used_len:]
            if pdu:
                if pdu.dev_id != dev_id:
                    raise ModbusIOException(
                        f"ERROR: request ask for id={dev_id} but id={pdu.dev_id}, CLOSING CONNECTION."
                    )
                return self.trace_pdu(False, pdu)

    def sync_execute(self, no_response_expected: bool, request: ModbusPDU) -> ModbusPDU:
        """Execute requests asynchronously.

        REMARK: this method is identical to execute, apart from the lock and sync_receive.
                any changes in either method MUST be mirrored !!!
        """
        if not self.sync_client.connect():
            raise ConnectionException("Client cannot connect (automatic retry continuing) !!")
        with self._sync_lock:
            request.transaction_id = self.getNextTID()
            count_retries = 0
            while count_retries <= self.retries:
                self.pdu_send(request)
                if no_response_expected:
                    return ExceptionResponse(0xff)
                try:
                    return self.sync_get_response(request.dev_id)
                except asyncio.exceptions.TimeoutError:
                    count_retries += 1
            if self.count_until_disconnect < 0:
                self.connection_lost(asyncio.TimeoutError("Server not responding"))
                raise ModbusIOException(
                    "ERROR: No response received of the last requests (default: retries+3), CLOSING CONNECTION."
                )
            self.count_until_disconnect -= 1
            txt = f"No response received after {self.retries} retries, continue with next request"
            Log.error(txt)
            raise ModbusIOException(txt)

    async def execute(self, no_response_expected: bool, request: ModbusPDU) -> ModbusPDU:
        """Execute requests asynchronously.

        REMARK: this method is identical to sync_execute, apart from the lock and try/except.
                any changes in either method MUST be mirrored !!!
        """
        if not self.transport:
            Log.warning("Not connected, trying to connect!")
            if not await self.connect():
                raise ConnectionException("Client cannot connect (automatic retry continuing) !!")
        async with self._lock:
            request.transaction_id = self.getNextTID()
            count_retries = 0
            while count_retries <= self.retries:
                self.recv_buffer = b""
                self.response_future = asyncio.Future()
                self.pdu_send(request)
                if no_response_expected:
                    return ExceptionResponse(0xff)
                try:
                    response = await asyncio.wait_for(
                        self.response_future, timeout=self.comm_params.timeout_connect
                    )
                    self.count_until_disconnect= self.max_until_disconnect
                    if response.dev_id != request.dev_id:
                        raise ModbusIOException(
                            f"ERROR: request ask for id={request.dev_id} but got id={response.dev_id}, CLOSING CONNECTION."
                        )
                    return response
                except asyncio.exceptions.TimeoutError:
                    count_retries += 1
                except asyncio.exceptions.CancelledError as exc:
                    raise ModbusIOException("Request cancelled outside pymodbus.") from exc
            if self.count_until_disconnect < 0:
                self.connection_lost(asyncio.TimeoutError("Server not responding"))
                raise ModbusIOException(
                    "ERROR: No response received of the last requests (default: retries+3), CLOSING CONNECTION."
                )
            self.count_until_disconnect -= 1
            txt = f"No response received after {self.retries} retries, continue with next request"
            Log.error(txt)
            raise ModbusIOException(txt)

    def pdu_send(self, pdu: ModbusPDU, addr: tuple | None = None) -> None:
        """Build byte stream and send."""
        self.request_dev_id = pdu.dev_id
        self.request_transaction_id = pdu.transaction_id
        packet = self.framer.buildFrame(self.trace_pdu(True, pdu))
        if self.is_sync and self.comm_params.handle_local_echo:
            self.sent_buffer = packet
        self.low_level_send(self.trace_packet(True, packet), addr=addr)

    def callback_new_connection(self):
        """Call when listener receive new connection request."""

    def callback_connected(self) -> None:
        """Call when connection is succcesfull."""
        self.count_until_disconnect = self.max_until_disconnect
        self.next_tid = 0
        self.trace_connect(True)

    def callback_disconnected(self, exc: Exception | None) -> None:
        """Call when connection is lost."""
        self.trace_connect(False)

    def callback_data(self, data: bytes, addr: tuple | None = None) -> int:
        """Handle received data."""
        self.last_pdu = self.last_addr = None
        used_len, pdu = self.framer.processIncomingFrame(self.trace_packet(False, data))
        if pdu:
            self.last_pdu = self.trace_pdu(False, pdu)
            self.last_addr = addr
            if not self.is_server:
                if pdu.dev_id != self.request_dev_id:
                    Log.warning(f"ERROR: expected id {self.request_dev_id} but got {pdu.dev_id}, IGNORING.")
                elif pdu.transaction_id != self.request_transaction_id:
                    Log.warning(f"ERROR: expected transaction {self.request_transaction_id} but got {pdu.transaction_id}, IGNORING.")
                elif self.response_future.done():
                    Log.warning("ERROR: received pdu without a corresponding request, IGNORING")
                else:
                    self.response_future.set_result(self.last_pdu)
        return used_len

    def getNextTID(self) -> int:
        """Retrieve the next transaction identifier."""
        if isinstance(self.framer, (FramerAscii, FramerRTU)):
          self.next_tid = 0
        elif self.next_tid >= 65000:
            self.next_tid = 1
        else:
            self.next_tid += 1
        return self.next_tid



# -*- coding: utf-8 -*-

"""package dmm4050
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2018
license   GPL v3.0+
brief     Dedicated to handle the Tektronix DMM4050 digital multimeter.
"""

import logging
import re
# import serial
import socket
from PyQt5.QtCore import QObject, pyqtSignal, pyqtSlot

from scinstr.dmm.constants import DMM4050_ID

# Multimeter functions
(CAP, CURR, CURRDC, VOLTAC, VOLTDC, VOLTDCRAT, RES, FRES, FRE, PER,
 TEMPRTD, TEMPFRTD, DIOD, CONT) = ("CAP", "CURR", "CURR:DC",
                                   "VOLT:AC", "VOLT:DC", "VOLT:DC:RAT",
                                   "RES", "FRES", "FRE", "PER",
                                   "TEMP:RTD", "TEMP:FRTD",
                                   "DIOD", "CONT")
# Integration time
(PLC002, PLC02, PLC1, PLC10, PLC100) = ("0.02", "0.2", "1", "10", "100")

# Voltage input range
(RANGE10MV, RANGE1V, RANGE10V, RANGE100V, RANGE1000V) = ("0.1", "1", "10",
                                                         "100", "1000")
# Trigger source
IMM, EXT, BUS = "IMM", "EXT", "BUS"


# =============================================================================
class Dmm4050Abstract(QObject):
    """Abstract class to handling DMM4050 digital multimeter device. Derived
    classes need to re-implement specific-protocol methods: connect(), close(),
    _write(), _read()...
    """

    connected = pyqtSignal()
    closed = pyqtSignal()
    id_checked = pyqtSignal(bool)
    out_updated = pyqtSignal((float,), (str,))

    def __init__(self):
        """The constructor.
        """
        super().__init__()

    @pyqtSlot()
    def connect(self):
        """Abstract protocol connect process. Derived classes must implement
        the connect process dedicated to the specific protocol used.
        :returns: None
        """
        pass

    @pyqtSlot()
    def close(self):
        """Abstract protocol closing process. Derived classes must implement
        the closing process dedicated to the specific protocol used.
        :returns: None
        """
        pass

    def _write(self, data):
        """Abstract protocol write process. Derived classes must implement
        the write process dedicated to the specific protocol used.
        :param data: data writes to device (str)
        :returns: None
        """
        pass

    def _read(self, length):
        """Abstract protocol read process. Derived classes must implement
        the read process dedicated to the specific protocol used.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        pass

    def write(self, data):
        """A basic write method: writes "data" to device.
        :param data: data writes to device (str)
        :returns: None
        """
        try:
            self._write(data + '\n')
        except Exception:
            pass  # Handled in _write() subclass

    def read(self, length):
        """A basic read method: read a message from device.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        try:
            retval = self._read(length)
        except Exception:
            pass  # Handled in _read() subclass
        else:
            return retval

    def reset(self):
        """Resets meter to its power-on state, sets all bits to zero in
        status byte register and all event registers and clear error queue.
        :returns: None
        """
        try:
            self._write("*RST")
            self._write("*CLS")
        except Exception as ex:
            logging.error(str(ex))
        else:
            logging.info("Reset meter")

    def dc_voltage_config(self, intg_time=None, rang=None, res=None,
                          d_fltr=None, a_fltr=None):
        """Configures specific parameters of DC voltage measurement.
        Note that you must specify a range whenever specifying a resolution
        parameter.
        :param intg_time: integration time in PLC (Power Line Cycle) (int)
        :param rang: DC voltage range (str)
        :param res: DC measurement resolution (int)
        :param d_fltr: Set digital averaging filter (bool)
        :param a_fltr: Set analog filter (bool)
        :returns: None
        """
        try:
            msg = "CONF:VOLT:DC"
            if rang is not None:
                msg += " " + rang
                if res is not None:
                    msg += ", " + str(res)
            self._write(msg)
            if a_fltr is not None:
                self._write("FILT:DC:STAT" + str(a_fltr))
            if intg_time < 1:
                self._write("FILT:DC:DIG:STAT OFF")
            elif d_fltr is not None:
                self._write("FILT:DC:DIG:STAT" + str(d_fltr))
            if intg_time is not None:
                self._write("VOLT:DC:NPLC " + str(intg_time))
        except Exception as ex:
            logging.error("%r", ex)
        else:
            logging.info("DC voltage measurement config done")

    def trigger_config(self, src=IMM, delay=None, delauto=None, scount=None,
                       tcount=None):
        """Configures trigger.
        :param src: source from which the Meter will expect a measurement
        trigger (str)
        :param delay: delay between receiving a trigger and the beginning of
        measurement cycle (int)
        :param delauto: enables or disables automatic trigger delay.
        The amount of delay is based on selected function, integration time,
        and filter setting. When automatic delay is enabled, "delay" parameters
        is ignored (int)
        :param scount: number of measurements the Meter takes per trigger
        (int or str)
        :param tcount: number of triggers the Meter will take before switching
        to an idle state. If you use "INF", continously accepts triggers,
        in this case use a device clear to stop the process (int or str)
        :returns: None
        """
        try:
            self._write("TRIG:SOUR " + src)
            if delauto == 1:
                self._write("TRIG:DEL:AUTO ON")
            elif delauto == 0:
                self._write("TRIG:DEL:AUTO OFF")
            elif delay is not None:
                self._write("TRIG:DEL " + str(delay))
            if tcount is not None:
                self._write("TRIG:COUN " + str(tcount))
            if scount is not None:
                self._write("SAMP:COUN " + str(scount))
        except Exception as ex:
            logging.error("%r", ex)
        else:
            logging.info("Trigger configuration done")

    def init_trig(self):
        """Once the Meter has been configured for a measurement, ini_meas()
        causes the Meter to take a measurement when the trigger conditions have
        been met. The measurement reading(s) are placed in the Meter’s internal
        memory: up to 5000 to be read at a later time using the FETCh? command.
        :returns: None
        """
        try:
            self._write("INIT")
        except Exception as ex:
            logging.error("%r", ex)
        else:
            logging.info("Initiate triggering")

    def data_read_fetch(self):
        """To process readings from the Meter’s internal memory to the output
        buffer.
        :returns: data read in internal memory buffer (float)
        """
        try:
            self._write("FETC?")
            retval = self._read(100)  # Note '100' not used in serial com mode
        except Exception as ex:
            logging.error(str(ex))
        else:
            logging.debug("Data read (fetch) measurement: " + retval)
        return float(retval)

    def data_read(self):
        """Takes a measurement the next time the trigger condition is met.
        After the measurement is taken, the reading is placed in the output
        buffer. "data_read" will not cause readings to be stored in the Meter’s
        internal memory.
        :returns: data read in buffer (float)
        """
        try:
            self._write("READ?")
            retval = self._read(100)  # Note '100' not used in serial com mode
            if re.match("[\+\-]\d\.\d*E[\+\-]\d\d", retval) is None:
                return None
        except Exception as ex:
            logging.error(str(ex))
            return 0.0  # Default value
        self.out_updated[str].emit(retval)
        self.out_updated[float].emit(float(retval))
        logging.debug("Data read measurement: " + retval)
        return float(retval)


# =============================================================================
class Dmm4050Eth(Dmm4050Abstract):
    """Class using to handle the DMM4050 device through ethernet protocol.
    """

    DEFAULT_IP = "192.168.0.32"
    DEFAULT_PORT = 3490
    DEFAULT_TIMEOUT = 0.5

    def __init__(self, ip=DEFAULT_IP, port=DEFAULT_PORT,
                 timeout=DEFAULT_TIMEOUT):
        """The constructor.
        :param ip: IP address of device (str)
        :param port: Ethernet port of device (int)
        :param timeout: Socket timeout value in s (float)
        :returns: None
        """
        super().__init__()
        self._sock = None
        self._ip = ip
        self._port = port
        self._timeout = timeout

    @pyqtSlot()
    def connect(self):
        """Specific ethernet connection process to DMM4050.
        :returns: True if connection success other False (Bool)
        """
        self._sock = socket.socket(socket.AF_INET,
                                   socket.SOCK_STREAM,
                                   socket.IPPROTO_TCP)
        self._sock.settimeout(self._timeout)
        try:
            self._sock.connect((self._ip, self._port))
        except ValueError as ex:
            logging.error("Connection parameters out of range: %r", ex)
            return False
        except socket.timeout:
            logging.error("Timeout on connection")
            return False
        except Exception as ex:
            logging.critical("Unexpected exception during connection with " +
                             "DMM4050: %r", ex)
            return False
        self.connected.emit()
        logging.info("Connected to DMM4050")
        return True

    @pyqtSlot()
    def close(self):
        """Specific ethernet closing process with DMM4050.
        :returns: True if connection success other False (Bool)
        """
        try:
            self._sock.close()
        except Exception as ex:
            print("dmm.close.exception")
            logging.error("%s", str(ex))
            return False
        self._sock = None
        self.closed.emit()
        logging.info("Connection to DMM4050 closed")
        return True

    @pyqtSlot()
    def check_interface(self):
        """Basic interface connection test: check id of device.
        Return True if interface with device is OK.
        :returns: status of interface with device (bool)
        """
        try:
            self.connect()
            _id = str(self.get_id())
        except Exception:  # Catch connection and get_id problem
            _id = ""  # To avoid error with 'find()' if '_id' is not defined
        else:
            self.close()
        if _id.find(DMM4050_ID) >= 0:
            self.id_checked.emit(True)
            return True
        else:
            self.id_checked.emit(False)
            return False

    def _write(self, data):
        """Specific ethernet writing process.
        :param data: data writes to device (str)
        :returns: None
        """
        try:
            self._sock.send(data.encode('utf8'))
        except socket.timeout:
            logging.error("DMM timeout")
        except Exception as ex:
            logging.error(str(ex))
            raise Exception(ex)
        else:
            logging.debug("_write " + data)

    def _read(self, length):
        """Specific ethernet reading process.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        try:
            retval = (self._sock.recv(length)).decode('utf-8')
        except socket.timeout:
            logging.error("DMM timeout")
        except Exception as ex:
            logging.error(str(ex))
        else:
            logging.debug("_read %r", retval)
            return retval

    @pyqtSlot()
    def get_id(self):
        """Return product id of device.
        :returns: product id of device (str)
        """
        self._write("*IDN?\n")
        _id = self._read(100)
        return _id

    @pyqtSlot(float)
    def set_timeout(self, timeout):
        """Sets timeout on socket operations.
        :param timeout: timeout value in second (float)
        :returns: None
        """
        self._sock.settimeout(timeout)

    @pyqtSlot()
    def get_timeout(self):
        """Gets timeout on socket operations.
        :returns: timeout value in second (float)
        """
        return self._sock.gettimeout()

    @pyqtSlot(str)
    def set_ip(self, ip):
        """Sets IP address used to speak with device.
        :param ip: IP address (str)
        :return: None
        """
        self._ip = ip

    @pyqtSlot()
    def get_ip(self):
        """Gets IP used to speak with device.
        :returns: IP address (str)
        """
        return self._ip

    @pyqtSlot(int)
    def set_port(self, port):
        """Sets internet port used to speak with device.
        :param port: port used by dmm4050 (int)
        :returns: None
        """
        self._port = port

    @pyqtSlot()
    def get_port(self):
        """Gets internet port used to speak with device.
        :returns: port used by dmm4050 (int)
        """
        return self._port


# =============================================================================
class Dmm4050Emul(Dmm4050Abstract):
    """Class using to emulate the DMM4050 device.
    """

    def __init__(self, *args):
        super().__init__()
        import random
        self._datas = [random.uniform(-1, 1) for _ in range(100)]
        # Add 'spikes'
        self._datas[43], self._datas[74], self._datas[75] = -35, 44, 45
        self._did = 0  # self._datas index
        logging.info("Dmm4050Emul initialized with args :" + str(args))

    @pyqtSlot()
    def check_interface(self):
        """Basic interface connection test: check id of device.
        Return True if interface with device is OK.
        :returns: status of interface with device (bool)
        """
        self.id_checked.emit(True)
        return True

    @pyqtSlot()
    def get_id(self):
        super().get_id()

    @pyqtSlot(float)
    def set_timeout(self, timeout):
        logging.debug("Set timeout to " + str(timeout))

    @pyqtSlot()
    def get_timeout(self):
        logging.debug("Get timeout")
        return 0.1

    @pyqtSlot(str)
    def set_ip(self, ip):
        logging.debug("Set ip to " + str(ip))

    @pyqtSlot()
    def get_ip(self):
        logging.debug("Get IP")
        return "192.168.0.2"

    @pyqtSlot(int)
    def set_port(self, port):
        """Sets internet port used to speak with device.
        :param port: port used by dmm4050 (int)
        :returns: None
        """
        logging.debug("Set port to " + str(port))

    @pyqtSlot()
    def get_port(self):
        logging.debug("Get port")
        return 1

    @pyqtSlot()
    def connect(self):
        """Abstract protocol connect process. Derived classes must implement
        the connect process dedicated to the specific protocol used.
        :returns: None
        """
        logging.info("Connected to DMM")
        self.connected.emit()
        return True

    @pyqtSlot()
    def close(self):
        """Abstract protocol closing process. Derived classes must implement
        the closing process dedicated to the specific protocol used.
        :returns: None
        """
        logging.info("Connection to DMM closed")
        self.closed.emit()
        return True

    def _write(self, data):
        """Abstract protocol write process. Derived classes must implement
        the write process dedicated to the specific protocol used.
        :param data: data writes to device (str)
        :returns: None
        """
        logging.debug("Write: " + data)

    def _read(self, length):
        """Abstract protocol read process. Derived classes must implement
        the read process dedicated to the specific protocol used.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        retval = self._datas[self._did]
        if self._did < len(self._datas) - 1:
            self._did += 1
        else:
            self._did = 0
        logging.info("Read: %E", retval)
        return "{:+E}".format(retval)

    def write(self, data):
        """A basic write method: writes "data" to device.
        :param data: data writes to device (str)
        :returns: None
        """
        super().write(data)

    def read(self, length):
        """A basic read method: read a message from device.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        super().read(length)

    def reset(self):
        """Resets meter to its power-on state, sets all bits to zero in
        status byte register and all event registers and clear error queue.
        :returns: None
        """
        super().reset()

    def dc_voltage_config(self, intg_time=None, rang=None, res=None,
                          d_fltr=None, a_fltr=None):
        """Configures specific parameters of DC voltage measurement.
        Note that you must specify a range whenever specifying a resolution
        parameter.
        :param intg_time: integration time in PLC (Power Line Cycle) (int)
        :param rang: DC voltage range (str)
        :param res: DC measurement resolution (int)
        :param d_fltr: Set digital averaging filter (bool)
        :param a_fltr: Set analog filter (bool)
        :returns: None
        """
        super().dc_voltage_config(intg_time, rang, res, d_fltr, a_fltr)

    def trigger_config(self, src=None, delay=None, delauto=None, scount=None,
                       tcount=None):
        """Configures trigger.
        :param src: source from which the Meter will expect a measurement
        trigger (str)
        :param delay: delay between receiving a trigger and the beginning of
        measurement cycle (int)
        :param delauto: enables or disables automatic trigger delay.
        The amount of delay is based on selected function, integration time,
        and filter setting. When automatic delay is enabled, "delay" parameters
        is ignored (int)
        :param scount: number of measurements the Meter takes per trigger
        (int or str)
        :param tcount: number of triggers the Meter will take before switching
        to an idle state. If you use "INF", continously accepts triggers,
        in this case use a device clear to stop the process (int or str)
        :returns: None
        """
        super().trigger_config(src, delay, delauto, scount, tcount)

    def init_trig(self):
        """Once the Meter has been configured for a measurement, ini_meas()
        causes the Meter to take a measurement when the trigger conditions have
        been met. The measurement reading(s) are placed in the Meter’s internal
        memory (up to 5000 to be read at a later time using the FETCh? command.
        :returns: None
        """
        super().init_trig()

    def data_read_fetch(self):
        """To process readings from the Meter’s internal memory to the output
        buffer.
        :returns: data read in internal memory buffer (float)
        """
        super().data_read_fetch()

    def data_read(self):
        """Takes a measurement the next time the trigger condition is met.
        After the measurement is taken, the reading is placed in the output
        buffer. "data_read" will not cause readings to be stored in the Meter’s
        internal memory.
        :returns: data read in buffer (float)
        """
        super().data_read()


# =============================================================================
def check_dmm4050():
    """Checks the Dmm4050xx class: connects to the multimeter, configures a dc
    voltage measurement then collects and prints data to standard output.
    Values that work:
    - 1PLC with 0.5 de rtimeout
    - 10PLC with 0.8 de rtimeout
    """
    from datetime import datetime

    date_fmt = "%d/%m/%Y %H:%M:%S"
    log_format = "%(asctime)s %(levelname) -8s %(filename)s " + \
                 " %(funcName)s (%(lineno)d): %(message)s"
    logging.basicConfig(level=logging.INFO,
                        datefmt=date_fmt,
                        format=log_format)

    # dmm4050 = Dmm4050Serial(serport="/dev/ttyUSB0", timeout=0.8)
    dmm4050 = Dmm4050Eth(ip="192.168.0.32", port=3490, timeout=0.8)
    dmm4050.connect()
    dmm4050.reset()
    dmm4050.write("ZERO:AUTO OFF")
    dmm4050.write("CONF:VOLT:DC DEF")  # Autorange
    dmm4050.write("ZERO:AUTO OFF")  # Autozero off
    dmm4050.write("VOLT:DC:NPLC 10")
    dmm4050.write("TRIG:SOUR IMM")  # Immediate trigger source
    dmm4050.write("TRIG:DEL 0")   # No trigger delay
    dmm4050.write("TRIG:COUN 1")  # 1 measurement per trigger
    dmm4050.write("SAMP:COUN 1")  # 1 sample per trigger
    dmm4050.write("SYST:REM\n")   # DMM in remote mode
    try:
        while True:
            value = dmm4050.data_read()
            now = datetime.utcnow()
            if value is None or value == "":
                print("# No data @", now)
            else:
                print(now, value)
    except KeyboardInterrupt:
        pass
    except Exception as er:
        print("# Exception during acquisition: " + str(er))
    dmm4050.reset()
    dmm4050.write("SYST:LOC\n")
    dmm4050.close()


# =============================================================================
if __name__ == '__main__':
    check_dmm4050()



"""Modbus Client Common."""
from __future__ import annotations

import struct
from abc import abstractmethod
from enum import Enum
from typing import Generic, Literal, TypeVar, cast

import pymodbus.pdu.bit_message as pdu_bit
import pymodbus.pdu.diag_message as pdu_diag
import pymodbus.pdu.file_message as pdu_file_msg
import pymodbus.pdu.mei_message as pdu_mei
import pymodbus.pdu.other_message as pdu_other_msg
import pymodbus.pdu.register_message as pdu_reg
from pymodbus.constants import ModbusStatus
from pymodbus.exceptions import ModbusException
from pymodbus.pdu.pdu import ModbusPDU, pack_bitstring, unpack_bitstring


T = TypeVar("T", covariant=False)


class ModbusClientMixin(Generic[T]):  # pylint: disable=too-many-public-methods
    """**ModbusClientMixin**.

    This is an interface class to facilitate the sending requests/receiving responses like read_coils.
    execute() allows to make a call with non-standard or user defined function codes (remember to add a PDU
    in the transport class to interpret the request/response).

    Simple modbus message call::

        response = client.read_coils(1, 10)
        # or
        response = await client.read_coils(1, 10)

    Advanced modbus message call::

        request = ReadCoilsRequest(1,10)
        response = client.execute(False, request)
        # or
        request = ReadCoilsRequest(1,10)
        response = await client.execute(False, request)

    .. tip::
        All methods can be used directly (synchronous) or
        with await <method> (asynchronous) depending on the client used.
    """

    def __init__(self):
        """Initialize."""

    @abstractmethod
    def execute(self, no_response_expected: bool, request: ModbusPDU) -> T:
        """Execute request."""

    def read_coils(self, address: int, *, count: int = 1, slave: int = 1, no_response_expected: bool = False) -> T:
        """Read coils (code 0x01).

        :param address: Start address to read from
        :param count: (optional) Number of coils to read
        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        reads from 1 to 2000 contiguous in a remote device (slave).

        Coils are addressed as 0-N (Note some device manuals uses 1-N, assuming 1==0).
        """
        return self.execute(no_response_expected, pdu_bit.ReadCoilsRequest(address=address, count=count, dev_id=slave))

    def read_discrete_inputs(self,
                             address: int,
                             *,
                             count: int = 1,
                             slave: int = 1,
                             no_response_expected: bool = False) -> T:
        """Read discrete inputs (code 0x02).

        :param address: Start address to read from
        :param count: (optional) Number of coils to read
        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        read from 1 to 2000(0x7d0) discrete inputs (bits) in a remote device.

        Discrete Inputs are addressed as 0-N (Note some device manuals uses 1-N, assuming 1==0).
        """
        pdu = pdu_bit.ReadDiscreteInputsRequest(address=address, count=count, dev_id=slave)
        return self.execute(no_response_expected, pdu)

    def read_holding_registers(self,
                               address: int,
                               *,
                               count: int = 1,
                               slave: int = 1,
                               no_response_expected: bool = False) -> T:
        """Read holding registers (code 0x03).

        :param address: Start address to read from
        :param count: (optional) Number of registers to read
        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function is used to read the contents of a contiguous block
        of holding registers in a remote device. The Request specifies the
        starting register address and the number of registers.

        Registers are addressed starting at zero.
        Therefore devices that specify 1-16 are addressed as 0-15.
        """
        return self.execute(no_response_expected, pdu_reg.ReadHoldingRegistersRequest(address=address, count=count, dev_id=slave))

    def read_input_registers(self,
                             address: int,
                             *,
                             count: int = 1,
                             slave: int = 1,
                             no_response_expected: bool = False) -> T:
        """Read input registers (code 0x04).

        :param address: Start address to read from
        :param count: (optional) Number of coils to read
        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function is used to read from 1 to approx. 125 contiguous
        input registers in a remote device. The Request specifies the
        starting register address and the number of registers.

        Registers are addressed starting at zero.
        Therefore devices that specify 1-16 are addressed as 0-15.
        """
        return self.execute(no_response_expected, pdu_reg.ReadInputRegistersRequest(address=address, count=count, dev_id=slave))

    def write_coil(self, address: int, value: bool, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Write single coil (code 0x05).

        :param address: Address to write to
        :param value: Boolean to write
        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        write ON/OFF to a single coil in a remote device.

        Coils are addressed as 0-N (Note some device manuals uses 1-N, assuming 1==0).
        """
        pdu = pdu_bit.WriteSingleCoilRequest(address=address, bits=[value], dev_id=slave)
        return self.execute(no_response_expected, pdu)

    def write_register(self, address: int, value: int, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Write register (code 0x06).

        :param address: Address to write to
        :param value: Value to write
        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function is used to write a single holding register in a remote device.

        The Request specifies the address of the register to be written.

        Registers are addressed starting at zero. Therefore register
        numbered 1 is addressed as 0.
        """
        return self.execute(no_response_expected, pdu_reg.WriteSingleRegisterRequest(address=address, registers=[value], dev_id=slave))

    def read_exception_status(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Read Exception Status (code 0x07).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function is used to read the contents of eight Exception Status outputs in a remote device.

        The function provides a simple method for
        accessing this information, because the Exception Output references are
        known (no output reference is needed in the function).
        """
        return self.execute(no_response_expected, pdu_other_msg.ReadExceptionStatusRequest(dev_id=slave))

    def diag_query_data(self, msg: bytes, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose query data (code 0x08 sub 0x00).

        :param msg: Message to be returned
        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        The data passed in the request data field is to be returned (looped back)
        in the response. The entire response message should be identical to the
        request.
        """
        return self.execute(no_response_expected, pdu_diag.ReturnQueryDataRequest(msg, dev_id=slave))

    def diag_restart_communication(self, toggle: bool, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose restart communication (code 0x08 sub 0x01).

        :param toggle: True if toggled.
        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        The remote device serial line port must be initialized and restarted, and
        all of its communications event counters are cleared. If the port is
        currently in Listen Only Mode, no response is returned. This function is
        the only one that brings the port out of Listen Only Mode. If the port is
        not currently in Listen Only Mode, a normal response is returned. This
        occurs before the restart is update_datastored.
        """
        msg = ModbusStatus.ON if toggle else ModbusStatus.OFF
        return self.execute(no_response_expected, pdu_diag.RestartCommunicationsOptionRequest(message=msg, dev_id=slave))

    def diag_read_diagnostic_register(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose read diagnostic register (code 0x08 sub 0x02).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        The contents of the remote device's 16-bit diagnostic register are returned in the response.
        """
        return self.execute(no_response_expected, pdu_diag.ReturnDiagnosticRegisterRequest(dev_id=slave))

    def diag_change_ascii_input_delimeter(self, *, delimiter: int = 0x0a, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose change ASCII input delimiter (code 0x08 sub 0x03).

        :param delimiter: char to replace LF
        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        The character passed in the request becomes the end of
        message delimiter for future messages (replacing the default LF
        character). This function is useful in cases of a Line Feed is not
        required at the end of ASCII messages.
        """
        return self.execute(no_response_expected, pdu_diag.ChangeAsciiInputDelimiterRequest(message=delimiter, dev_id=slave))

    def diag_force_listen_only(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose force listen only (code 0x08 sub 0x04).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        Forces the addressed remote device to its Listen Only Mode for MODBUS communications.

        This isolates it from the other devices on the network,
        allowing them to continue communicating without interruption from the
        addressed remote device. No response is returned.
        """
        return self.execute(no_response_expected, pdu_diag.ForceListenOnlyModeRequest(dev_id=slave))

    def diag_clear_counters(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose clear counters (code 0x08 sub 0x0A).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        Clear ll counters and the diagnostic register. Also, counters are cleared upon power-up
        """
        return self.execute(no_response_expected, pdu_diag.ClearCountersRequest(dev_id=slave))

    def diag_read_bus_message_count(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose read bus message count (code 0x08 sub 0x0B).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        The response data field returns the quantity of messages that the
        remote device has detected on the communications systems since its last
        restart, clear counters operation, or power-up
        """
        return self.execute(no_response_expected, pdu_diag.ReturnBusMessageCountRequest(dev_id=slave))

    def diag_read_bus_comm_error_count(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose read Bus Communication Error Count (code 0x08 sub 0x0C).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        The response data field returns the quantity of CRC errors encountered
        by the remote device since its last restart, clear counter operation, or
        power-up
        """
        return self.execute(no_response_expected, pdu_diag.ReturnBusCommunicationErrorCountRequest(dev_id=slave))

    def diag_read_bus_exception_error_count(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose read Bus Exception Error Count (code 0x08 sub 0x0D).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        The response data field returns the quantity of modbus exception
        responses returned by the remote device since its last restart,
        clear counters operation, or power-up
        """
        return self.execute(no_response_expected, pdu_diag.ReturnBusExceptionErrorCountRequest(dev_id=slave))

    def diag_read_slave_message_count(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose read Slave Message Count (code 0x08 sub 0x0E).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        The response data field returns the quantity of messages addressed to the
        remote device, that the remote device has processed since
        its last restart, clear counters operation, or power-up
        """
        return self.execute(no_response_expected, pdu_diag.ReturnSlaveMessageCountRequest(dev_id=slave))

    def diag_read_slave_no_response_count(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose read Slave No Response Count (code 0x08 sub 0x0F).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        The response data field returns the quantity of messages addressed to the
        remote device, that the remote device has processed since
        its last restart, clear counters operation, or power-up.
        """
        return self.execute(no_response_expected, pdu_diag.ReturnSlaveNoResponseCountRequest(dev_id=slave))

    def diag_read_slave_nak_count(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose read Slave NAK Count (code 0x08 sub 0x10).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        The response data field returns the quantity of messages addressed to the
        remote device for which it returned a Negative ACKNOWLEDGE (NAK) exception
        response, since its last restart, clear counters operation, or power-up.
        Exception responses are described and listed in section 7 .
        """
        return self.execute(no_response_expected, pdu_diag.ReturnSlaveNAKCountRequest(dev_id=slave))

    def diag_read_slave_busy_count(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose read Slave Busy Count (code 0x08 sub 0x11).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        The response data field returns the quantity of messages addressed to the
        remote device for which it returned a Slave Device Busy exception response,
        since its last restart, clear counters operation, or power-up.
        """
        return self.execute(no_response_expected, pdu_diag.ReturnSlaveBusyCountRequest(dev_id=slave))

    def diag_read_bus_char_overrun_count(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose read Bus Character Overrun Count (code 0x08 sub 0x12).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        The response data field returns the quantity of messages addressed to the
        remote device that it could not handle due to a character overrun condition,
        since its last restart, clear counters operation, or power-up. A character
        overrun is caused by data characters arriving at the port faster than they
        can be stored, or by the loss of a character due to a hardware malfunction.
        """
        return self.execute(no_response_expected, pdu_diag.ReturnSlaveBusCharacterOverrunCountRequest(dev_id=slave))

    def diag_read_iop_overrun_count(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose read Iop overrun count (code 0x08 sub 0x13).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        An IOP overrun is caused by data characters arriving at the port
        faster than they can be stored, or by the loss of a character due
        to a hardware malfunction.  This function is specific to the 884.
        """
        return self.execute(no_response_expected, pdu_diag.ReturnIopOverrunCountRequest(dev_id=slave))

    def diag_clear_overrun_counter(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose Clear Overrun Counter and Flag (code 0x08 sub 0x14).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        An error flag should be cleared, but nothing else in the
        specification mentions is, so it is ignored.
        """
        return self.execute(no_response_expected, pdu_diag.ClearOverrunCountRequest(dev_id=slave))

    def diag_getclear_modbus_response(self, *, data: int = 0, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose Get/Clear modbus plus (code 0x08 sub 0x15).

        :param data: "Get Statistics" or "Clear Statistics"
        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        In addition to the Function code (08) and Subfunction code
        (00 15 hex) in the query, a two-byte Operation field is used
        to specify either a "Get Statistics" or a "Clear Statistics"
        operation.  The two operations are exclusive - the "Get"
        operation cannot clear the statistics, and the "Clear"
        operation does not return statistics prior to clearing
        them. Statistics are also cleared on power-up of the slave
        device.
        """
        return self.execute(no_response_expected, pdu_diag.GetClearModbusPlusRequest(message=data, dev_id=slave))

    def diag_get_comm_event_counter(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose get event counter (code 0x0B).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function is used to get a status word and an event count from the remote device.

        By fetching the current count before and after a series of messages, a
        client can determine whether the messages were handled normally by the
        remote device.

        The device's event counter is incremented once for each successful
        message completion. It is not incremented for exception responses,
        poll commands, or fetch event counter commands.

        The event counter can be reset by means of the Diagnostics function
        Restart Communications or Clear Counters and Diagnostic Register.
        """
        return self.execute(no_response_expected, pdu_other_msg.GetCommEventCounterRequest(dev_id=slave))

    def diag_get_comm_event_log(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Diagnose get event counter (code 0x0C).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function is used to get a status word.

        Event count, message count, and a field of event bytes from the remote device.

        The status word and event counts are identical to that returned by
        the Get Communications Event Counter function.

        The message counter contains the quantity of messages processed by the
        remote device since its last restart, clear counters operation, or
        power-up. This count is identical to that returned by the Diagnostic
        function Return Bus Message Count.

        The event bytes field contains 0-64 bytes, with each byte corresponding
        to the status of one MODBUS send or receive operation for the remote
        device. The remote device enters the events into the field in
        chronological order. Byte 0 is the most recent event. Each new byte
        flushes the oldest byte from the field.
        """
        return self.execute(no_response_expected, pdu_other_msg.GetCommEventLogRequest(dev_id=slave))

    def write_coils(
        self,
        address: int,
        values: list[bool],
        *,
        slave: int = 1,
        no_response_expected: bool = False
    ) -> T:
        """Write coils (code 0x0F).

        :param address: Start address to write to
        :param values: List of booleans to write, or a single boolean to write
        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        write ON/OFF to multiple coils in a remote device.

        Coils are addressed as 0-N (Note some device manuals uses 1-N, assuming 1==0).
        """
        pdu = pdu_bit.WriteMultipleCoilsRequest(address=address, bits=values, dev_id=slave)
        return self.execute(no_response_expected, pdu)

    def write_registers(
        self,
        address: int,
        values: list[int],
        *,
        slave: int = 1,
        no_response_expected: bool = False
    ) -> T:
        """Write registers (code 0x10).

        :param address: Start address to write to
        :param values: List of values to write
        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function is used to write a block of contiguous registers
        (1 to approx. 120 registers) in a remote device.
        """
        return self.execute(no_response_expected, pdu_reg.WriteMultipleRegistersRequest(address=address, registers=values,dev_id=slave))

    def report_slave_id(self, *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Report slave ID (code 0x11).

        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function is used to read the description of the type, the current status
        and other information specific to a remote device.
        """
        return self.execute(no_response_expected, pdu_other_msg.ReportSlaveIdRequest(dev_id=slave))

    def read_file_record(self, records: list[pdu_file_msg.FileRecord], *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Read file record (code 0x14).

        :param records: List of FileRecord (Reference type, File number, Record Number)
        :param slave: device id
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function is used to perform a file record read. All request
        data lengths are provided in terms of number of bytes and all record
        lengths are provided in terms of registers.

        A file is an organization of records. Each file contains 10000 records,
        addressed 0000 to 9999 decimal or 0x0000 to 0x270f. For example, record
        12 is addressed as 12. The function can read multiple groups of
        references. The groups can be separating (non-contiguous), but the
        references within each group must be sequential. Each group is defined
        in a separate "sub-request" field that contains seven bytes::

            The reference type: 1 byte
            The file number: 2 bytes
            The starting record number within the file: 2 bytes
            The length of the record to be read: 2 bytes

        The quantity of registers to be read, combined with all other fields
        in the expected response, must not exceed the allowable length of the
        MODBUS PDU: 235 bytes.
        """
        return self.execute(no_response_expected, pdu_file_msg.ReadFileRecordRequest(records, dev_id=slave))

    def write_file_record(self, records: list[pdu_file_msg.FileRecord], *, slave: int = 1, no_response_expected: bool = False) -> T:
        """Write file record (code 0x15).

        :param records: List of File_record (Reference type, File number, Record Number, Record Length, Record Data)
        :param slave: (optional) Device id
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function is used to perform a file record write. All
        request data lengths are provided in terms of number of bytes
        and all record lengths are provided in terms of the number of 16
        bit words.
        """
        return self.execute(no_response_expected, pdu_file_msg.WriteFileRecordRequest(records=records, dev_id=slave))

    def mask_write_register(
        self,
        *,
        address: int = 0x0000,
        and_mask: int = 0xFFFF,
        or_mask: int = 0x0000,
        slave: int = 1,
        no_response_expected: bool = False
    ) -> T:
        """Mask write register (code 0x16).

        :param address: The mask pointer address (0x0000 to 0xffff)
        :param and_mask: The and bitmask to apply to the register address
        :param or_mask: The or bitmask to apply to the register address
        :param slave: (optional) device id
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function is used to modify the contents of a specified holding register
        using a combination of an AND mask, an OR mask, and the register's current contents.

        The function can be used to set or clear individual bits in the register.
        """
        return self.execute(no_response_expected, pdu_reg.MaskWriteRegisterRequest(address=address, and_mask=and_mask, or_mask=or_mask, dev_id=slave))

    def readwrite_registers(
        self,
        *,
        read_address: int = 0,
        read_count: int = 0,
        write_address: int = 0,
        address: int | None = None,
        values: list[int] | None = None,
        slave: int = 1,
        no_response_expected: bool = False
    ) -> T:
        """Read/Write registers (code 0x17).

        :param read_address: The address to start reading from
        :param read_count: The number of registers to read from address
        :param write_address: The address to start writing to
        :param address: (optional) use as read/write address
        :param values: List of values to write, or a single value to write
        :param slave: (optional) Modbus slave ID
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function performs a combination of one read operation and one
        write operation in a single MODBUS transaction. The write
        operation is performed before the read.

        Holding registers are addressed starting at zero. Therefore holding
        registers 1-16 are addressed in the PDU as 0-15.
        """
        if not values:
            values = []
        if address:
            read_address = address
            write_address = address
        return self.execute(no_response_expected, pdu_reg.ReadWriteMultipleRegistersRequest( read_address=read_address, read_count=read_count, write_address=write_address, write_registers=values,dev_id=slave))

    def read_fifo_queue(self, *, address: int = 0x0000, slave: int = 1, no_response_expected: bool = False) -> T:
        """Read FIFO queue (code 0x18).

        :param address: The address to start reading from
        :param slave: (optional) device id
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function allows to read the contents of a First-In-First-Out
        (FIFO) queue of register in a remote device. The function returns a
        count of the registers in the queue, followed by the queued data.
        Up to 32 registers can be read: the count, plus up to 31 queued data
        registers.

        The queue count register is returned first, followed by the queued data
        registers.  The function reads the queue contents, but does not clear
        them.
        """
        return self.execute(no_response_expected, pdu_file_msg.ReadFifoQueueRequest(address, dev_id=slave))

    # code 0x2B sub 0x0D: CANopen General Reference Request and Response, NOT IMPLEMENTED

    def read_device_information(self, *, read_code: int | None = None,
                                object_id: int = 0x00,
                                slave: int = 1,
                                no_response_expected: bool = False) -> T:
        """Read FIFO queue (code 0x2B sub 0x0E).

        :param read_code: The device information read code
        :param object_id: The object to read from
        :param slave: (optional) Device id
        :param no_response_expected: (optional) The client will not expect a response to the request
        :raises ModbusException:

        This function allows reading the identification and additional
        information relative to the physical and functional description of a
        remote device, only.

        The Read Device Identification interface is modeled as an address space
        composed of a set of addressable data elements. The data elements are
        called objects and an object Id identifies them.
        """
        return self.execute(no_response_expected, pdu_mei.ReadDeviceInformationRequest(read_code, object_id, dev_id=slave))

    # ------------------
    # Converter methods
    # ------------------

    class DATATYPE(Enum):
        """Datatype enum (name and internal data), used for convert_* calls."""

        INT16 = ("h", 1)
        UINT16 = ("H", 1)
        INT32 = ("i", 2)
        UINT32 = ("I", 2)
        INT64 = ("q", 4)
        UINT64 = ("Q", 4)
        FLOAT32 = ("f", 2)
        FLOAT64 = ("d", 4)
        STRING = ("s", 0)
        BITS = ("bits", 0)

    @classmethod
    def convert_from_registers(
        cls, registers: list[int], data_type: DATATYPE, word_order: Literal["big", "little"] = "big", string_encoding: str = "utf-8"
    ) -> int | float | str | list[bool] | list[int] | list[float]:
        """Convert registers to int/float/str.

        :param registers: list of registers received from e.g. read_holding_registers()
        :param data_type: data type to convert to
        :param word_order: "big"/"little" order of words/registers
        :param string_encoding: The encoding with which to decode the bytearray, only used when data_type=DATATYPE.STRING
        :returns: scalar or array of "data_type"
        :raises ModbusException: when size of registers is not a multiple of data_type
        :raises ParameterException: when the specified string encoding is not supported
        """
        if not (data_len := data_type.value[1]):
            byte_list = bytearray()
            if word_order == "little":
                registers.reverse()
            for x in registers:
                byte_list.extend(int.to_bytes(x, 2, "big"))
            if data_type == cls.DATATYPE.STRING:
                trailing_nulls_begin = len(byte_list)
                while trailing_nulls_begin > 0 and not byte_list[trailing_nulls_begin - 1]:
                    trailing_nulls_begin -= 1
                byte_list = byte_list[:trailing_nulls_begin]
                return byte_list.decode(string_encoding)
            return unpack_bitstring(byte_list)
        if (reg_len := len(registers)) % data_len:
            raise ModbusException(
                f"Registers illegal size ({len(registers)}) expected multiple of {data_len}!"
            )

        result = []
        for i in range(0, reg_len, data_len):
            regs = registers[i:i+data_len]
            if word_order == "little":
                regs.reverse()
            byte_list = bytearray()
            for x in regs:
                byte_list.extend(int.to_bytes(x, 2, "big"))
            result.append(struct.unpack(f">{data_type.value[0]}", byte_list)[0])
        return result if len(result) != 1 else result[0]

    @classmethod
    def convert_to_registers(
        cls, value: int | float | str | list[bool] | list[int] | list[float], data_type: DATATYPE, word_order: Literal["big", "little"] = "big", string_encoding: str = "utf-8"
    ) -> list[int]:
        """Convert int/float/str to registers (16/32/64 bit).

        :param value: value to be converted
        :param data_type: data type to convert from
        :param word_order: "big"/"little" order of words/registers
        :param string_encoding: The encoding with which to encode the bytearray, only used when data_type=DATATYPE.STRING
        :returns: List of registers, can be used directly in e.g. write_registers()
        :raises TypeError: when there is a mismatch between data_type and value
        :raises ParameterException: when the specified string encoding is not supported
        """
        if data_type == cls.DATATYPE.BITS:
            if not isinstance(value, list):
                raise TypeError(f"Value should be list of bool but is {type(value)}.")
            if (missing := len(value) % 16):
                value = value + [False] * (16 - missing)
            byte_list = pack_bitstring(cast(list[bool], value))
        elif data_type == cls.DATATYPE.STRING:
            if not isinstance(value, str):
                raise TypeError(f"Value should be string but is {type(value)}.")
            byte_list = value.encode(string_encoding)
            if len(byte_list) % 2:
                byte_list += b"\x00"
        else:
            if not isinstance(value, list):
                value = cast(list[int], [value])
            byte_list = bytearray()
            for v in value:
                byte_list.extend(struct.pack(f">{data_type.value[0]}", v))
        regs = [
            int.from_bytes(byte_list[x : x + 2], "big")
            for x in range(0, len(byte_list), 2)
        ]
        if word_order == "little":
            regs.reverse()
        return regs



# -*- coding: utf-8 -*-

""" scinstr/daq/labjack/t7/constants.py """

ORGANIZATION = "FEMTO_Engineering"
APP_BRIEF = "Library dedicated to handle the Labjack DAQ board without LJM lib"
AUTHOR_NAME = "Benoit Dubois"
AUTHOR_MAIL = "benoit.dubois@femto-engineering.fr"
COPYRIGHT = "FEMTO Engineering"
LICENSE = "GNU GPL v3.0 or upper."

T7_ID = "7"

# Default T7 ethernet parameters
T7_PORT = 502
T7_STREAM_PORT = 702
T7_ETH_TIMEOUT = 0.5  # in s

# Default T7 USB parameters
T7_VID = 0x0CD5
T7_PID = 0x0007
T7_USB_TIMEOUT = 0  # in ms



# -*- coding: utf-8 -*-

import serial
from scinstr.vacuum.gauge import GaugeSerial, CR, LF

LINE_TERMINATION = CR + LF # CR, LF and CRLF are all possible

ERR_CODES = {
    0: 'No error',
    1: 'Watchdog has responded',
    2: 'Task fail error',
    5: 'EPROM error',
    6: 'RAM error',
    7: 'EEPROM error',
    9: 'DISPLAY error',
    10: 'A/D converter error',
    11: 'Gauge error (e.g. filament rupture, no supply)',
    12: 'Gauge identification error',
}

PRESSURE_READING_STATUS = {
    0: 'Measurement data okay',
    1: 'Underrange',
    2: 'Overrange',
    3: 'Sensor error',
    4: 'Sensor off',
    5: 'No sensor',
    6: 'Identification error',
    7: 'Error ITR'
}

GAUGE_TYPE = {
    'TTR': 'Pirani, all version',
    'TTR100': 'Pirani/Capacitive',
    'PTR': 'Cold Cathode',
    'PTR90': 'Cold Cathode/Pirani',
    'CTR': 'Capacitive',
    'ITR': 'Hot cathode',
    'ITR100': 'Hot cathode',
    'ITR200': 'Hot cathode/Pirani',
    'noSen': 'no Sensor',
    'noid': 'no identifier'
}

FSR = {
    0: '0.01 mbar',
    1: '0.01 Torr',
    2: '0.02 Torr',
    3: '0.05 Torr',
    4: '0.10 mbar',
    5: '0.10 Torr',
    6: '0.25 mbar',
    7: '0.25 Torr',
    8: '0.50 mbar',
    9: '0.50 Torr',
    10: '1 mbar',
    11: '1 Torr',
    12: '2 mbar',
    13: '2 Torr',
    14: '5 mbar',
    15: '5 Torr',
    16: '10 mbar',
    17: '10 Torr',
    18: '20 mbar',
    19: '20 Torr',
    20: '50 mbar',
    21: '50 Torr',
    22: '100 mbar',
    23: '100 Torr',
    24: '200 mbar',
    25: '200 Torr',
    26: '500 mbar',
    27: '500 Torr',
    28: '1000 mbar',
    29: '1100 mbar',
    30: '1000 Torr',
    31: '2 bar',
    32: '5 bar',
    33: '10 bar',
    34: '50 bar',
    35: 'DI200 mbar',
    36: 'DI200 bar',
    37: 'DI200 barRel'
}

SER_TIMEOUT = 0.2
ETH_TIMEOUT = 0.2


# =============================================================================
class CenteroneSerial(GaugeSerial):
    """Handle Centerone gauge device through serial interface.
    Serial interface configuration parameters:
    - Baud rate 9600, 19200, 38400
    - Start bits 1
    - Data bits 8
    - Parity None
    - Stop bits 1
    - Flow control None
    - Handshaking None
    """

    def __init__(self, port, timeout=SER_TIMEOUT):
        super().__init__(port,
                         baudrate=9600,
                         bytesize=8,
                         parity=serial.PARITY_NONE,
                         stopbits=1,
                         timeout=timeout)
